@@ -20,30 +20,37 @@
 #include "xfrd.h"
 #include "xfrd-notify.h"
 #include "difffile.h"
+#include "rrl.h"
 
 /* attempt to send NSD_STATS command to child fd */
 static void send_stat_to_child(struct main_ipc_handler_data* data, int fd);
 /* send reload request over the IPC channel */
-static void xfrd_send_reload_req(xfrd_state_t* xfrd);
+static void xfrd_send_reload_req(xfrd_state_type* xfrd);
 /* send quit request over the IPC channel */
-static void xfrd_send_quit_req(xfrd_state_t* xfrd);
+static void xfrd_send_quit_req(xfrd_state_type* xfrd);
 /* perform read part of handle ipc for xfrd */
-static void xfrd_handle_ipc_read(struct event* handler, xfrd_state_t* xfrd);
+static void xfrd_handle_ipc_read(struct event* handler, xfrd_state_type* xfrd);
+static void ipc_child_quit(struct nsd* nsd) ATTR_NORETURN;
 
 static void
 ipc_child_quit(struct nsd* nsd)
 {
 	/* call shutdown and quit routines */
 	nsd->mode = NSD_QUIT;
+	service_remaining_tcp(nsd);
 #ifdef	BIND8_STATS
 	bind8_stats(nsd);
 #endif /* BIND8_STATS */
 
-#if 0 /* OS collects memory pages */
-	event_base_free(event_base);
-	region_destroy(server_region);
+#ifdef MEMCLEAN /* OS collects memory pages */
+#ifdef RATELIMIT
+	rrl_deinit(nsd->this_child->child_num);
+#endif
+	event_base_free(nsd->event_base);
+	region_destroy(nsd->server_region);
 #endif
 	server_shutdown(nsd);
+	/* ENOTREACH */
 	exit(0);
 }
 
@@ -82,21 +89,10 @@ child_handle_parent_command(int fd, short event, void* arg)
 		server_close_all_sockets(data->nsd->udp, data->nsd->ifs);
 		server_close_all_sockets(data->nsd->tcp, data->nsd->ifs);
 		/* mode == NSD_QUIT_CHILD */
-		(void)write(fd, &mode, sizeof(mode));
-		ipc_child_quit(data->nsd);
-		break;
-	case NSD_QUIT_WITH_STATS:
-#ifdef BIND8_STATS
-		DEBUG(DEBUG_IPC, 2, (LOG_INFO, "quit QUIT_WITH_STATS"));
-		/* reply with ack and stats and then quit */
-		if(!write_socket(fd, &mode, sizeof(mode))) {
-			log_msg(LOG_ERR, "cannot write quitwst to parent");
+		if(write(fd, &mode, sizeof(mode)) == -1) {
+			VERBOSITY(3, (LOG_INFO, "quit child write: %s",
+				strerror(errno)));
 		}
-		if(!write_socket(fd, &data->nsd->st, sizeof(data->nsd->st))) {
-			log_msg(LOG_ERR, "cannot write stats to parent");
-		}
-		fsync(fd);
-#endif /* BIND8_STATS */
 		ipc_child_quit(data->nsd);
 		break;
 	default:
@@ -170,39 +166,10 @@ send_stat_to_child(struct main_ipc_handler_data* data, int fd)
 	data->child->need_to_send_STATS = 0;
 }
 
-#ifndef NDEBUG
-int packet_read_query_section(buffer_type *packet, uint8_t* dest, uint16_t* qtype, uint16_t* qclass);
-static void
-debug_print_fwd_name(int ATTR_UNUSED(len), buffer_type* packet, int acl_num)
-{
-	uint8_t qnamebuf[MAXDOMAINLEN];
-	uint16_t qtype, qclass;
-	const dname_type* dname;
-	region_type* tempregion = region_create(xalloc, free);
-
-	size_t bufpos = buffer_position(packet);
-	buffer_rewind(packet);
-	buffer_skip(packet, 12);
-	if(packet_read_query_section(packet, qnamebuf, &qtype, &qclass)) {
-		dname = dname_make(tempregion, qnamebuf, 1);
-		log_msg(LOG_INFO, "main: fwd packet for %s, acl %d",
-			dname_to_string(dname,0), acl_num);
-	} else {
-		log_msg(LOG_INFO, "main: fwd packet badqname, acl %d", acl_num);
-	}
-	buffer_set_position(packet, bufpos);
-	region_destroy(tempregion);
-}
-#endif
-
 static void
 send_quit_to_child(struct main_ipc_handler_data* data, int fd)
 {
-#ifdef BIND8_STATS
-	sig_atomic_t cmd = NSD_QUIT_WITH_STATS;
-#else
 	sig_atomic_t cmd = NSD_QUIT;
-#endif
 	if(write(fd, &cmd, sizeof(cmd)) == -1) {
 		if(errno == EAGAIN || errno == EINTR)
 			return; /* try again later */
@@ -224,10 +191,23 @@ child_is_done(struct nsd* nsd, int fd)
 	for(i=0; i<nsd->child_count; ++i)
 		if(nsd->children[i].child_fd == fd) {
 			nsd->children[i].child_fd = -1;
-			nsd->children[i].has_exited = 1;
 			nsd->children[i].handler->fd = -1;
-			DEBUG(DEBUG_IPC,1, (LOG_INFO, "server %d is done",
-				(int)nsd->children[i].pid));
+			if(nsd->children[i].need_to_exit) {
+				DEBUG(DEBUG_IPC,1, (LOG_INFO, "server %d is done",
+					(int)nsd->children[i].pid));
+				nsd->children[i].has_exited = 1;
+			} else {
+				log_msg(LOG_WARNING,
+				       "server %d died unexpectedly, restarting",
+				       (int)nsd->children[i].pid);
+				/* this child is now going to be re-forked as
+				 * a subprocess of this server-main, and if a
+				 * reload is in progress the other children
+				 * are subprocesses of reload.  Until the
+				 * reload is done and they are all reforked. */
+				nsd->children[i].pid = -1;
+				nsd->restart_children = 1;
+			}
 		}
 	parent_check_all_children_exited(nsd);
 }
@@ -238,17 +218,19 @@ void
 stats_add(struct nsdst* total, struct nsdst* s)
 {
 	unsigned i;
-	for(i=0; i<sizeof(total->qtype)/sizeof(stc_t); i++)
+	for(i=0; i<sizeof(total->qtype)/sizeof(stc_type); i++)
 		total->qtype[i] += s->qtype[i];
-	for(i=0; i<sizeof(total->qclass)/sizeof(stc_t); i++)
+	for(i=0; i<sizeof(total->qclass)/sizeof(stc_type); i++)
 		total->qclass[i] += s->qclass[i];
 	total->qudp += s->qudp;
 	total->qudp6 += s->qudp6;
 	total->ctcp += s->ctcp;
 	total->ctcp6 += s->ctcp6;
-	for(i=0; i<sizeof(total->rcode)/sizeof(stc_t); i++)
+	total->ctls += s->ctls;
+	total->ctls6 += s->ctls6;
+	for(i=0; i<sizeof(total->rcode)/sizeof(stc_type); i++)
 		total->rcode[i] += s->rcode[i];
-	for(i=0; i<sizeof(total->opcode)/sizeof(stc_t); i++)
+	for(i=0; i<sizeof(total->opcode)/sizeof(stc_type); i++)
 		total->opcode[i] += s->opcode[i];
 	total->dropped += s->dropped;
 	total->truncated += s->truncated;
@@ -259,28 +241,41 @@ stats_add(struct nsdst* total, struct nsdst* s)
 	total->ednserr += s->ednserr;
 	total->raxfr += s->raxfr;
 	total->nona += s->nona;
+	total->rixfr += s->rixfr;
 
 	total->db_disk = s->db_disk;
 	total->db_mem = s->db_mem;
 }
 
-#define FINAL_STATS_TIMEOUT 10 /* seconds */
-static void
-read_child_stats(struct nsd* nsd, struct nsd_child* child, int fd)
+/** subtract stats from total */
+void
+stats_subtract(struct nsdst* total, struct nsdst* s)
 {
-	struct nsdst s;
-	errno=0;
-	if(block_read(nsd, fd, &s, sizeof(s), FINAL_STATS_TIMEOUT)!=sizeof(s)) {
-		log_msg(LOG_ERR, "problems reading finalstats from server "
-			"%d: %s", (int)child->pid, strerror(errno));
-	} else {
-		stats_add(&nsd->st, &s);
-		child->query_count = s.qudp + s.qudp6 + s.ctcp + s.ctcp6;
-		/* we know that the child is going to close the connection
-		 * now (this is an ACK of the QUIT_W_STATS so we know the
-		 * child is done, no longer sending e.g. NOTIFY contents) */
-		child_is_done(nsd, fd);
-	}
+	unsigned i;
+	for(i=0; i<sizeof(total->qtype)/sizeof(stc_type); i++)
+		total->qtype[i] -= s->qtype[i];
+	for(i=0; i<sizeof(total->qclass)/sizeof(stc_type); i++)
+		total->qclass[i] -= s->qclass[i];
+	total->qudp -= s->qudp;
+	total->qudp6 -= s->qudp6;
+	total->ctcp -= s->ctcp;
+	total->ctcp6 -= s->ctcp6;
+	total->ctls -= s->ctls;
+	total->ctls6 -= s->ctls6;
+	for(i=0; i<sizeof(total->rcode)/sizeof(stc_type); i++)
+		total->rcode[i] -= s->rcode[i];
+	for(i=0; i<sizeof(total->opcode)/sizeof(stc_type); i++)
+		total->opcode[i] -= s->opcode[i];
+	total->dropped -= s->dropped;
+	total->truncated -= s->truncated;
+	total->wrongzone -= s->wrongzone;
+	total->txerr -= s->txerr;
+	total->rxerr -= s->rxerr;
+	total->edns -= s->edns;
+	total->ednserr -= s->ednserr;
+	total->raxfr -= s->raxfr;
+	total->nona -= s->nona;
+	total->rixfr -= s->rixfr;
 }
 #endif /* BIND8_STATS */
 
@@ -295,7 +290,7 @@ parent_handle_child_command(netio_type *ATTR_UNUSED(netio),
 		(struct main_ipc_handler_data*)handler->user_data;
 
 	/* do a nonblocking write to the child if it is ready. */
-	if (event_types & NETIO_EVENT_WRITE) {
+	if ((event_types & NETIO_EVENT_WRITE)) {
 		if(data->child->need_to_send_STATS &&
 			!data->child->need_to_exit) {
 			send_stat_to_child(data, handler->fd);
@@ -312,95 +307,6 @@ parent_handle_child_command(netio_type *ATTR_UNUSED(netio),
 		return;
 	}
 
-	if (data->forward_mode) {
-		int got_acl;
-		/* forward the data to xfrd */
-		DEBUG(DEBUG_IPC,2, (LOG_INFO,
-			"main passed packet readup %d", (int)data->got_bytes));
-		if(data->got_bytes < sizeof(data->total_bytes))
-		{
-			if ((len = read(handler->fd,
-				(char*)&data->total_bytes+data->got_bytes,
-				sizeof(data->total_bytes)-data->got_bytes)) == -1) {
-				log_msg(LOG_ERR, "handle_child_command: read: %s",
-					strerror(errno));
-				return;
-			}
-			if(len == 0) {
-				/* EOF */
-				data->forward_mode = 0;
-				return;
-			}
-			data->got_bytes += len;
-			if(data->got_bytes < sizeof(data->total_bytes))
-				return;
-			data->total_bytes = ntohs(data->total_bytes);
-			buffer_clear(data->packet);
-			if(data->total_bytes > buffer_capacity(data->packet)) {
-				log_msg(LOG_ERR, "internal error: ipc too large");
-				exit(1);
-			}
-			return;
-		}
-		/* read the packet */
-		if(data->got_bytes-sizeof(data->total_bytes) < data->total_bytes) {
-			if((len = read(handler->fd, buffer_current(data->packet),
-				data->total_bytes - (data->got_bytes-sizeof(data->total_bytes))
-				)) == -1 ) {
-				log_msg(LOG_ERR, "handle_child_command: read: %s",
-					strerror(errno));
-				return;
-			}
-			if(len == 0) {
-				/* EOF */
-				data->forward_mode = 0;
-				return;
-			}
-			data->got_bytes += len;
-			buffer_skip(data->packet, len);
-			/* read rest later */
-			return;
-		}
-		/* read the acl numbers */
-		got_acl = data->got_bytes - sizeof(data->total_bytes) - data->total_bytes;
-		if((len = read(handler->fd, (char*)&data->acl_num+got_acl,
-			sizeof(data->acl_num)+sizeof(data->acl_xfr)-got_acl)) == -1 ) {
-			log_msg(LOG_ERR, "handle_child_command: read: %s",
-				strerror(errno));
-			return;
-		}
-		if(len == 0) {
-			/* EOF */
-			data->forward_mode = 0;
-			return;
-		}
-		got_acl += len;
-		data->got_bytes += len;
-		if(got_acl >= (int)(sizeof(data->acl_num)+sizeof(data->acl_xfr))) {
-			uint16_t len = htons(data->total_bytes);
-			DEBUG(DEBUG_IPC,2, (LOG_INFO,
-				"main fwd passed packet write %d", (int)data->got_bytes));
-#ifndef NDEBUG
-			if(nsd_debug_level >= 2)
-				debug_print_fwd_name(len, data->packet, data->acl_num);
-#endif
-			data->forward_mode = 0;
-			mode = NSD_PASS_TO_XFRD;
-			if(!write_socket(*data->xfrd_sock, &mode, sizeof(mode)) ||
-			   !write_socket(*data->xfrd_sock, &len, sizeof(len)) ||
-			   !write_socket(*data->xfrd_sock, buffer_begin(data->packet),
-				data->total_bytes) ||
-			   !write_socket(*data->xfrd_sock, &data->acl_num,
-			   	sizeof(data->acl_num)) ||
-			   !write_socket(*data->xfrd_sock, &data->acl_xfr,
-			   	sizeof(data->acl_xfr))) {
-				log_msg(LOG_ERR, "error in ipc fwd main2xfrd: %s",
-					strerror(errno));
-			}
-		}
-		return;
-	}
-
 	/* read command from ipc */
 	if ((len = read(handler->fd, &mode, sizeof(mode))) == -1) {
 		log_msg(LOG_ERR, "handle_child_command: read: %s",
@@ -417,23 +323,12 @@ parent_handle_child_command(netio_type *ATTR_UNUSED(netio),
 	case NSD_QUIT:
 		data->nsd->mode = mode;
 		break;
-#ifdef BIND8_STATS
-	case NSD_QUIT_WITH_STATS:
-		read_child_stats(data->nsd, data->child, handler->fd);
-		break;
-#endif /* BIND8_STATS */
 	case NSD_STATS:
 		data->nsd->signal_hint_stats = 1;
 		break;
 	case NSD_REAP_CHILDREN:
 		data->nsd->signal_hint_child = 1;
 		break;
-	case NSD_PASS_TO_XFRD:
-		/* set mode for handle_child_command; echo to xfrd. */
-		data->forward_mode = 1;
-		data->got_bytes = 0;
-		data->total_bytes = 0;
-		break;
 	default:
 		log_msg(LOG_ERR, "handle_child_command: bad mode %d",
 			(int) mode);
@@ -475,11 +370,12 @@ parent_handle_reload_command(netio_type *ATTR_UNUSED(netio),
 	}
 	if (len == 0)
 	{
-		if(handler->fd != -1) {
-			close(handler->fd);
-			handler->fd = -1;
-		}
+		assert(handler->fd != -1); /* or read() would have failed */
+		close(handler->fd);
+		handler->fd = -1;
+
 		log_msg(LOG_ERR, "handle_reload_cmd: reload closed cmd channel");
+		nsd->reload_failed = 1;
 		return;
 	}
 	switch (mode) {
@@ -508,7 +404,7 @@ parent_handle_reload_command(netio_type *ATTR_UNUSED(netio),
 }
 
 static void
-xfrd_send_reload_req(xfrd_state_t* xfrd)
+xfrd_send_reload_req(xfrd_state_type* xfrd)
 {
 	sig_atomic_t req = NSD_RELOAD;
 	uint64_t p = xfrd->last_task->data;
@@ -530,8 +426,8 @@ xfrd_send_reload_req(xfrd_state_t* xfrd)
 	task_remap(xfrd->nsd->task[xfrd->nsd->mytask]);
 	udb_ptr_init(xfrd->last_task, xfrd->nsd->task[xfrd->nsd->mytask]);
 	assert(udb_base_get_userdata(xfrd->nsd->task[xfrd->nsd->mytask])->data == 0);
-
-	xfrd_prepare_zones_for_reload();
+	if(!xfrd->reload_cmd_first_sent)
+		xfrd->reload_cmd_first_sent = xfrd_time();
 	xfrd->reload_cmd_last_sent = xfrd_time();
 	xfrd->need_to_send_reload = 0;
 	xfrd->can_send_reload = 0;
@@ -543,6 +439,7 @@ ipc_xfrd_set_listening(struct xfrd_state* xfrd, short mode)
 	int fd = xfrd->ipc_handler.ev_fd;
 	struct event_base* base = xfrd->event_base;
 	event_del(&xfrd->ipc_handler);
+	memset(&xfrd->ipc_handler, 0, sizeof(xfrd->ipc_handler));
 	event_set(&xfrd->ipc_handler, fd, mode, xfrd_handle_ipc, xfrd);
 	if(event_base_set(base, &xfrd->ipc_handler) != 0)
 		log_msg(LOG_ERR, "ipc: cannot set event_base");
@@ -553,7 +450,7 @@ ipc_xfrd_set_listening(struct xfrd_state* xfrd, short mode)
 }
 
 static void
-xfrd_send_shutdown_req(xfrd_state_t* xfrd)
+xfrd_send_shutdown_req(xfrd_state_type* xfrd)
 {
 	sig_atomic_t cmd = NSD_SHUTDOWN;
 	xfrd->ipc_send_blocked = 1;
@@ -567,7 +464,7 @@ xfrd_send_shutdown_req(xfrd_state_t* xfrd)
 }
 
 static void
-xfrd_send_quit_req(xfrd_state_t* xfrd)
+xfrd_send_quit_req(xfrd_state_type* xfrd)
 {
 	sig_atomic_t cmd = NSD_QUIT;
 	xfrd->ipc_send_blocked = 1;
@@ -581,7 +478,7 @@ xfrd_send_quit_req(xfrd_state_t* xfrd)
 }
 
 static void
-xfrd_send_stats(xfrd_state_t* xfrd)
+xfrd_send_stats(xfrd_state_type* xfrd)
 {
 	sig_atomic_t cmd = NSD_STATS;
 	DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: ipc send stats"));
@@ -595,14 +492,14 @@ xfrd_send_stats(xfrd_state_t* xfrd)
 void
 xfrd_handle_ipc(int ATTR_UNUSED(fd), short event, void* arg)
 {
-	xfrd_state_t* xfrd = (xfrd_state_t*)arg;
-        if ((event & EV_READ))
+	xfrd_state_type* xfrd = (xfrd_state_type*)arg;
+	if ((event & EV_READ))
 	{
 		/* first attempt to read as a signal from main
 		 * could block further send operations */
 		xfrd_handle_ipc_read(&xfrd->ipc_handler, xfrd);
 	}
-        if ((event & EV_WRITE))
+	if ((event & EV_WRITE))
 	{
 		if(xfrd->ipc_send_blocked) { /* wait for RELOAD_DONE */
 			ipc_xfrd_set_listening(xfrd, EV_PERSIST|EV_READ);
@@ -629,86 +526,42 @@ xfrd_handle_ipc(int ATTR_UNUSED(fd), short event, void* arg)
 }
 
 static void
-xfrd_handle_ipc_read(struct event* handler, xfrd_state_t* xfrd)
+xfrd_handle_ipc_read(struct event* handler, xfrd_state_type* xfrd)
 {
-        sig_atomic_t cmd;
-        int len;
+	sig_atomic_t cmd;
+	int len;
 
-	if(xfrd->ipc_conn->is_reading==2) {
-		buffer_type* tmp = xfrd->ipc_pass;
-		uint32_t acl_num;
-		int32_t acl_xfr;
-		/* read acl_num */
-		int ret = conn_read(xfrd->ipc_conn);
-		if(ret == -1) {
-			log_msg(LOG_ERR, "xfrd: error in read ipc: %s", strerror(errno));
-			xfrd->ipc_conn->is_reading = 0;
-			return;
-		}
-		if(ret == 0)
-			return;
-		buffer_flip(xfrd->ipc_conn->packet);
-		xfrd->ipc_pass = xfrd->ipc_conn->packet;
-		xfrd->ipc_conn->packet = tmp;
-		xfrd->ipc_conn->is_reading = 0;
-		acl_num = buffer_read_u32(xfrd->ipc_pass);
-		acl_xfr = (int32_t)buffer_read_u32(xfrd->ipc_pass);
-		xfrd_handle_passed_packet(xfrd->ipc_conn->packet, acl_num, acl_xfr);
-		return;
-	}
-	if(xfrd->ipc_conn->is_reading) {
-		/* reading an IPC message */
-		buffer_type* tmp;
-		int ret = conn_read(xfrd->ipc_conn);
-		if(ret == -1) {
-			log_msg(LOG_ERR, "xfrd: error in read ipc: %s", strerror(errno));
-			xfrd->ipc_conn->is_reading = 0;
-			return;
-		}
-		if(ret == 0)
-			return;
-		buffer_flip(xfrd->ipc_conn->packet);
-		/* use ipc_conn to read remaining data as well */
-		tmp = xfrd->ipc_pass;
-		xfrd->ipc_conn->is_reading=2;
-		xfrd->ipc_pass = xfrd->ipc_conn->packet;
-		xfrd->ipc_conn->packet = tmp;
-		xfrd->ipc_conn->total_bytes = sizeof(xfrd->ipc_conn->msglen);
-		xfrd->ipc_conn->msglen = 2*sizeof(uint32_t);
-		buffer_clear(xfrd->ipc_conn->packet);
-		buffer_set_limit(xfrd->ipc_conn->packet, xfrd->ipc_conn->msglen);
-		return;
-	}
-
-        if((len = read(handler->ev_fd, &cmd, sizeof(cmd))) == -1) {
+	if((len = read(handler->ev_fd, &cmd, sizeof(cmd))) == -1) {
 		if(errno != EINTR && errno != EAGAIN)
-                	log_msg(LOG_ERR, "xfrd_handle_ipc: read: %s",
-                        	strerror(errno));
-                return;
-        }
-        if(len == 0)
-        {
+			log_msg(LOG_ERR, "xfrd_handle_ipc: read: %s",
+				strerror(errno));
+		return;
+	}
+	if(len == 0)
+	{
 		/* parent closed the connection. Quit */
 		DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: main closed connection."));
 		xfrd->shutdown = 1;
 		return;
-        }
+	}
 
-        switch(cmd) {
-        case NSD_QUIT:
-        case NSD_SHUTDOWN:
+	switch(cmd) {
+	case NSD_QUIT:
+	case NSD_SHUTDOWN:
 		DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: main sent shutdown cmd."));
-                xfrd->shutdown = 1;
-                break;
+		xfrd->shutdown = 1;
+		break;
+	case NSD_RELOAD_FAILED:
+		xfrd->reload_failed = 1;
+		/* fall through */
 	case NSD_RELOAD_DONE:
 		/* reload has finished */
-		DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: ipc recv RELOAD_DONE"));
-#ifdef BIND8_STATS
+		DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: ipc recv %s",
+			xfrd->reload_failed ? "RELOAD FAILED" : "RELOAD DONE"));
 		if(block_read(NULL, handler->ev_fd, &xfrd->reload_pid,
 			sizeof(pid_t), -1) != sizeof(pid_t)) {
 			log_msg(LOG_ERR, "xfrd cannot get reload_pid");
 		}
-#endif /* BIND8_STATS */
 		/* read the not-mytask for the results and soainfo */
 		xfrd_process_task_result(xfrd,
 			xfrd->nsd->task[1-xfrd->nsd->mytask]);
@@ -718,11 +571,15 @@ xfrd_handle_ipc_read(struct event* handler, xfrd_state_t* xfrd)
 		xfrd->ipc_send_blocked = 0;
 		ipc_xfrd_set_listening(xfrd, EV_PERSIST|EV_READ|EV_WRITE);
 		xfrd_reopen_logfile();
-		xfrd_check_failed_updates();
-		break;
-	case NSD_PASS_TO_XFRD:
-		DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: ipc recv PASS_TO_XFRD"));
-		xfrd->ipc_conn->is_reading = 1;
+		if(!xfrd->reload_failed) {
+			xfrd_check_failed_updates();
+			xfrd->reload_cmd_first_sent = 0;
+		} else {
+			/* make reload happen again, right away */
+			xfrd_set_reload_now(xfrd);
+		}
+		xfrd_prepare_zones_for_reload();
+		xfrd->reload_failed = 0;
 		break;
 	case NSD_RELOAD_REQ:
 		DEBUG(DEBUG_IPC,1, (LOG_INFO, "xfrd: ipc recv RELOAD_REQ"));
@@ -737,16 +594,43 @@ xfrd_handle_ipc_read(struct event* handler, xfrd_state_t* xfrd)
 		ipc_xfrd_set_listening(xfrd, EV_PERSIST|EV_READ|EV_WRITE);
 		xfrd->need_to_send_quit = 1;
 		break;
-        default:
-                log_msg(LOG_ERR, "xfrd_handle_ipc: bad mode %d (%d)", (int)cmd,
+	default:
+		log_msg(LOG_ERR, "xfrd_handle_ipc: bad mode %d (%d)", (int)cmd,
 			(int)ntohl(cmd));
-                break;
-        }
-
-	if(xfrd->ipc_conn->is_reading) {
-		/* setup read of info */
-		xfrd->ipc_conn->total_bytes = 0;
-		xfrd->ipc_conn->msglen = 0;
-		buffer_clear(xfrd->ipc_conn->packet);
+		break;
 	}
 }
+
+void
+xfrd_handle_notify(int ATTR_UNUSED(fd), short event, void* arg)
+{
+	struct xfrd_tcp* notify_pipe = (struct xfrd_tcp*)arg;
+	uint32_t acl_num;
+	int32_t acl_xfr;
+
+	if(!(event & EV_READ))
+		return;
+
+	switch(conn_read(notify_pipe)){
+	case -1: /* TODO: What to do here? */
+		 return;
+	case  0: return; /* call back later */
+	default: break;
+	}
+	if(buffer_limit(notify_pipe->packet) < sizeof(acl_xfr)+sizeof(acl_num))
+		log_msg(LOG_ERR, "xfrd_handle_notify invalid message size");
+	else {
+		size_t eop = buffer_position(notify_pipe->packet)
+		           - sizeof(acl_xfr) - sizeof(acl_num);
+
+		buffer_set_position(notify_pipe->packet, eop);
+		acl_num = buffer_read_u32(notify_pipe->packet);
+		acl_xfr = (int32_t)buffer_read_u32(notify_pipe->packet);
+		buffer_set_position(notify_pipe->packet, eop);
+		buffer_flip(notify_pipe->packet);
+		xfrd_handle_passed_packet(notify_pipe->packet,acl_num,acl_xfr);
+	}
+	notify_pipe->total_bytes = 0;
+	notify_pipe->msglen = 0;
+	buffer_clear(notify_pipe->packet);
+}
