@@ -33,7 +33,12 @@ allocate_domain_info(domain_table_type* table,
 
 	result = (domain_type *) region_alloc(table->region,
 					      sizeof(domain_type));
-	result->dname = dname_partial_copy(
+#ifdef USE_RADIX_TREE
+	result->dname 
+#else
+	result->node.key
+#endif
+		= dname_partial_copy(
 		table->region, dname, domain_dname(parent)->label_count + 1);
 	result->parent = parent;
 	result->wildcard_child_closest_match = result;
@@ -68,23 +73,18 @@ allocate_domain_nsec3(domain_table_type* table, domain_type* result)
 	result->nsec3->nsec3_ds_parent_cover = NULL;
 	result->nsec3->nsec3_is_exact = 0;
 	result->nsec3->nsec3_ds_parent_is_exact = 0;
-	result->nsec3->have_nsec3_hash = 0;
-	result->nsec3->have_nsec3_wc_hash = 0;
-	result->nsec3->have_nsec3_ds_parent_hash = 0;
+	result->nsec3->hash_wc = NULL;
+	result->nsec3->ds_parent_hash = NULL;
 	result->nsec3->prehash_prev = NULL;
 	result->nsec3->prehash_next = NULL;
 	result->nsec3->nsec3_node.key = NULL;
-	result->nsec3->hash_node.key = NULL;
-	result->nsec3->wchash_node.key = NULL;
-	result->nsec3->dshash_node.key = NULL;
 }
 #endif /* NSEC3 */
 
-/** make the domain last in the numlist, changes numbers of domains */
-static void
+void
 numlist_make_last(domain_table_type* table, domain_type* domain)
 {
-	size_t sw;
+	uint32_t sw;
 	domain_type* last = table->numlist_last;
 	if(domain == last)
 		return;
@@ -121,8 +121,7 @@ numlist_make_last(domain_table_type* table, domain_type* domain)
 	table->numlist_last = domain;
 }
 
-/** pop the biggest domain off the numlist */
-static domain_type*
+domain_type*
 numlist_pop_last(domain_table_type* table)
 {
 	domain_type* d = table->numlist_last;
@@ -159,9 +158,9 @@ int domain_is_prehash(domain_table_type* table, domain_type* domain)
 
 /** remove domain node from NSEC3 tree in hash space */
 void
-zone_del_domain_in_hash_tree(rbtree_t* tree, rbnode_t* node)
+zone_del_domain_in_hash_tree(rbtree_type* tree, rbnode_type* node)
 {
-	if(!node->key)
+	if(!node->key || !tree)
 		return;
 	rbtree_delete(tree, node->key);
 	/* note that domain is no longer in the tree */
@@ -230,15 +229,27 @@ do_deldomain(namedb_type* db, domain_type* domain)
 		if(domain->nsec3->nsec3_node.key)
 			zone_del_domain_in_hash_tree(nsec3_tree_zone(db, domain)
 				->nsec3tree, &domain->nsec3->nsec3_node);
-		if(domain->nsec3->hash_node.key)
-			zone_del_domain_in_hash_tree(nsec3_tree_zone(db, domain)
-				->hashtree, &domain->nsec3->hash_node);
-		if(domain->nsec3->wchash_node.key)
-			zone_del_domain_in_hash_tree(nsec3_tree_zone(db, domain)
-				->wchashtree, &domain->nsec3->wchash_node);
-		if(domain->nsec3->dshash_node.key)
+		if(domain->nsec3->hash_wc) {
+			if(domain->nsec3->hash_wc->hash.node.key)
+				zone_del_domain_in_hash_tree(nsec3_tree_zone(db, domain)
+					->hashtree, &domain->nsec3->hash_wc->hash.node);
+			if(domain->nsec3->hash_wc->wc.node.key)
+				zone_del_domain_in_hash_tree(nsec3_tree_zone(db, domain)
+					->wchashtree, &domain->nsec3->hash_wc->wc.node);
+		}
+		if(domain->nsec3->ds_parent_hash && domain->nsec3->ds_parent_hash->node.key)
 			zone_del_domain_in_hash_tree(nsec3_tree_dszone(db, domain)
-				->dshashtree, &domain->nsec3->dshash_node);
+				->dshashtree, &domain->nsec3->ds_parent_hash->node);
+		if(domain->nsec3->hash_wc) {
+			region_recycle(db->domains->region,
+				domain->nsec3->hash_wc,
+				sizeof(nsec3_hash_wc_node_type));
+		}
+		if(domain->nsec3->ds_parent_hash) {
+			region_recycle(db->domains->region,
+				domain->nsec3->ds_parent_hash,
+				sizeof(nsec3_hash_node_type));
+		}
 		region_recycle(db->domains->region, domain->nsec3,
 			sizeof(struct nsec3_domain_data));
 	}
@@ -249,54 +260,49 @@ do_deldomain(namedb_type* db, domain_type* domain)
 	 * one-smaller than this domain as closest-match. */
 	if(domain->parent->wildcard_child_closest_match == domain)
 		domain->parent->wildcard_child_closest_match =
-			domain_previous(domain);
+			domain_previous_existing_child(domain);
 
 	/* actual removal */
+#ifdef USE_RADIX_TREE
 	radix_delete(db->domains->nametree, domain->rnode);
-	region_recycle(db->domains->region, (dname_type*)domain->dname,
-		dname_total_size(domain->dname));
+#else
+	rbtree_delete(db->domains->names_to_domains, domain->node.key);
+#endif
+	region_recycle(db->domains->region, domain_dname(domain),
+		dname_total_size(domain_dname(domain)));
 	region_recycle(db->domains->region, domain, sizeof(domain_type));
 }
 
 void
 domain_table_deldomain(namedb_type* db, domain_type* domain)
 {
+	domain_type* parent;
+
 	while(domain_can_be_deleted(domain)) {
+		parent = domain->parent;
 		/* delete it */
 		do_deldomain(db, domain);
 		/* test parent */
-		domain = domain->parent;
+		domain = parent;
 	}
 }
 
-/** clear hash tree */
-void
-hash_tree_clear(rbtree_t* tree)
+void hash_tree_delete(region_type* region, rbtree_type* tree)
 {
-	rbnode_t* n;
-	if(!tree) return;
-
-	/* note that elements are no longer in the tree */
-	for(n=rbtree_first(tree); n!=RBTREE_NULL; n=rbtree_next(n)) {
-		n->key = NULL;
-	}
-	tree->count = 0;
-	tree->root = RBTREE_NULL;
-}
-
-void hash_tree_delete(region_type* region, rbtree_t* tree)
-{
-	region_recycle(region, tree, sizeof(rbtree_t));
+	region_recycle(region, tree, sizeof(rbtree_type));
 }
 
 /** add domain nsec3 node to hashedspace tree */
-void zone_add_domain_in_hash_tree(region_type* region, rbtree_t** tree,
+void zone_add_domain_in_hash_tree(region_type* region, rbtree_type** tree,
 	int (*cmpf)(const void*, const void*),
-	domain_type* domain, rbnode_t* node)
+	domain_type* domain, rbnode_type* node)
 {
 	if(!*tree)
 		*tree = rbtree_create(region, cmpf);
-	memset(node, 0, sizeof(rbnode_t));
+	if(node->key && node->key == domain
+	&& rbtree_search(*tree, domain) == node)
+		return;
+	memset(node, 0, sizeof(rbnode_type));
 	node->key = domain;
 	rbtree_insert(*tree, node);
 }
@@ -313,7 +319,12 @@ domain_table_create(region_type* region)
 	origin = dname_make(region, (uint8_t *) "", 0);
 
 	root = (domain_type *) region_alloc(region, sizeof(domain_type));
-	root->dname = origin;
+#ifdef USE_RADIX_TREE
+	root->dname
+#else
+	root->node.key
+#endif
+		= origin;
 	root->parent = NULL;
 	root->wildcard_child_closest_match = root;
 	root->rrsets = NULL;
@@ -330,9 +341,15 @@ domain_table_create(region_type* region)
 	result = (domain_table_type *) region_alloc(region,
 						    sizeof(domain_table_type));
 	result->region = region;
+#ifdef USE_RADIX_TREE
 	result->nametree = radix_tree_create(region);
 	root->rnode = radname_insert(result->nametree, dname_name(root->dname),
 		root->dname->name_size, root);
+#else
+	result->names_to_domains = rbtree_create(
+		region, (int (*)(const void *, const void *)) dname_compare);
+	rbtree_insert(result->names_to_domains, (rbnode_type *) root);
+#endif
 
 	result->root = root;
 	result->numlist_last = root;
@@ -357,9 +374,13 @@ domain_table_search(domain_table_type *table,
 	assert(closest_match);
 	assert(closest_encloser);
 
+#ifdef USE_RADIX_TREE
 	exact = radname_find_less_equal(table->nametree, dname_name(dname),
 		dname->name_size, (struct radnode**)closest_match);
 	*closest_match = (domain_type*)((*(struct radnode**)closest_match)->elem);
+#else
+	exact = rbtree_find_less_equal(table->names_to_domains, dname, (rbnode_type **) closest_match);
+#endif
 	assert(*closest_match);
 
 	*closest_encloser = *closest_match;
@@ -416,9 +437,13 @@ domain_table_insert(domain_table_type* table,
 			result = allocate_domain_info(table,
 						      dname,
 						      closest_encloser);
+#ifdef USE_RADIX_TREE
 			result->rnode = radname_insert(table->nametree,
 				dname_name(result->dname),
 				result->dname->name_size, result);
+#else
+			rbtree_insert(table->names_to_domains, (rbnode_type *) result);
+#endif
 
 			/*
 			 * If the newly added domain name is larger
@@ -443,20 +468,18 @@ domain_table_insert(domain_table_type* table,
 	return result;
 }
 
-int
-domain_table_iterate(domain_table_type* table,
-		    domain_table_iterator_type iterator,
-		    void* user_data)
+domain_type *domain_previous_existing_child(domain_type* domain)
 {
-	int error = 0;
-	struct radnode* n;
-	for(n = radix_first(table->nametree); n; n = radix_next(n)) {
-		error += iterator((domain_type*)n->elem, user_data);
+	domain_type* parent = domain->parent;
+	domain = domain_previous(domain);
+	while(domain && !domain->is_existing) {
+		if(domain == parent) /* do not walk back above parent */
+			return parent;
+		domain = domain_previous(domain);
 	}
-	return error;
+	return domain;
 }
 
-
 void
 domain_add_rrset(domain_type* domain, rrset_type* rrset)
 {
@@ -474,6 +497,15 @@ domain_add_rrset(domain_type* domain, rrset_type* rrset)
 
 	while (domain && !domain->is_existing) {
 		domain->is_existing = 1;
+		/* does this name in existance update the parent's
+		 * wildcard closest match? */
+		if(domain->parent
+		   && label_compare(dname_name(domain_dname(domain)),
+			(const uint8_t *) "\001*") <= 0
+		   && dname_compare(domain_dname(domain),
+		   	domain_dname(domain->parent->wildcard_child_closest_match)) > 0) {
+			domain->parent->wildcard_child_closest_match = domain;
+		}
 		domain = domain->parent;
 	}
 }
@@ -507,15 +539,36 @@ domain_find_any_rrset(domain_type* domain, zone_type* zone)
 	return NULL;
 }
 
+rrset_type *
+domain_find_rrset_and_prev(domain_type* domain, zone_type* zone, uint16_t type,
+	rrset_type** prev)
+{
+	rrset_type* result = domain->rrsets, *prevp = NULL;
+
+	while (result) {
+		if (result->zone == zone && rrset_rrtype(result) == type) {
+			*prev = prevp;
+			return result;
+		}
+		prevp = result;
+		result = result->next;
+	}
+	*prev = NULL;
+	return NULL;
+}
+
 zone_type *
-domain_find_zone(domain_type* domain)
+domain_find_zone(namedb_type* db, domain_type* domain)
 {
 	rrset_type* rrset;
 	while (domain) {
-		for (rrset = domain->rrsets; rrset; rrset = rrset->next) {
-			if (rrset_rrtype(rrset) == TYPE_SOA) {
-				return rrset->zone;
+		if(domain->is_apex) {
+			for (rrset = domain->rrsets; rrset; rrset = rrset->next) {
+				if (rrset_rrtype(rrset) == TYPE_SOA) {
+					return rrset->zone;
+				}
 			}
+			return namedb_find_zone(db, domain_dname(domain));
 		}
 		domain = domain->parent;
 	}
@@ -523,7 +576,7 @@ domain_find_zone(domain_type* domain)
 }
 
 zone_type *
-domain_find_parent_zone(zone_type* zone)
+domain_find_parent_zone(namedb_type* db, zone_type* zone)
 {
 	rrset_type* rrset;
 
@@ -534,23 +587,47 @@ domain_find_parent_zone(zone_type* zone)
 			return rrset->zone;
 		}
 	}
+	/* the NS record in the parent zone above this zone is not present,
+	 * workaround to find that parent zone anyway */
+	if(zone->apex->parent)
+		return domain_find_zone(db, zone->apex->parent);
 	return NULL;
 }
 
 domain_type *
 domain_find_ns_rrsets(domain_type* domain, zone_type* zone, rrset_type **ns)
 {
+	/* return highest NS RRset in the zone that is a delegation above */
+	domain_type* result = NULL;
+	rrset_type* rrset = NULL;
 	while (domain && domain != zone->apex) {
-		*ns = domain_find_rrset(domain, zone, TYPE_NS);
-		if (*ns)
-			return domain;
+		rrset = domain_find_rrset(domain, zone, TYPE_NS);
+		if (rrset) {
+			*ns = rrset;
+			result = domain;
+		}
 		domain = domain->parent;
 	}
 
+	if(result)
+		return result;
+
 	*ns = NULL;
 	return NULL;
 }
 
+domain_type *
+find_dname_above(domain_type* domain, zone_type* zone)
+{
+	domain_type* d = domain->parent;
+	while(d && d != zone->apex) {
+		if(domain_find_rrset(d, zone, TYPE_DNAME))
+			return d;
+		d = d->parent;
+	}
+	return NULL;
+}
+
 int
 domain_is_glue(domain_type* domain, zone_type* zone)
 {
@@ -588,11 +665,11 @@ zone_is_secure(zone_type* zone)
 uint16_t
 rr_rrsig_type_covered(rr_type* rr)
 {
+	uint16_t type;
 	assert(rr->type == TYPE_RRSIG);
-	assert(rr->rdata_count > 0);
-	assert(rdata_atom_size(rr->rdatas[0]) == sizeof(uint16_t));
-
-	return ntohs(* (uint16_t *) rdata_atom_data(rr->rdatas[0]));
+	assert(rr->rdlength > 2);
+	memcpy(&type, rr->rdata, sizeof(type));
+	return ntohs(type);
 }
 
 zone_type *
@@ -635,3 +712,65 @@ namedb_lookup(struct namedb* db,
 	return domain_table_search(
 		db->domains, dname, closest_match, closest_encloser);
 }
+
+void zone_rr_iter_init(struct zone_rr_iter *iter, struct zone *zone)
+{
+	assert(iter != NULL);
+	assert(zone != NULL);
+	memset(iter, 0, sizeof(*iter));
+	iter->zone = zone;
+}
+
+rr_type *zone_rr_iter_next(struct zone_rr_iter *iter)
+{
+	assert(iter != NULL);
+	assert(iter->zone != NULL);
+
+	if(iter->index == -1) {
+		assert(iter->domain == NULL);
+		assert(iter->rrset == NULL);
+		return NULL;
+	} else if(iter->rrset == NULL) {
+		/* ensure SOA RR is returned first */
+		assert(iter->domain == NULL);
+		assert(iter->index == 0);
+		iter->rrset = iter->zone->soa_rrset;
+	}
+
+	while(iter->rrset != NULL) {
+		if(iter->index < iter->rrset->rr_count) {
+			return iter->rrset->rrs[iter->index++];
+		}
+		iter->index = 0;
+		if(iter->domain == NULL) {
+			assert(iter->rrset == iter->zone->soa_rrset);
+			iter->domain = iter->zone->apex;
+			iter->rrset = iter->domain->rrsets;
+		} else {
+			iter->rrset = iter->rrset->next;
+		}
+		/* ensure SOA RR is not returned again and RR belongs to zone */
+		while((iter->rrset == NULL && iter->domain != NULL) ||
+		      (iter->rrset != NULL && (iter->rrset == iter->zone->soa_rrset ||
+		                               iter->rrset->zone != iter->zone)))
+		{
+			if(iter->rrset != NULL) {
+				iter->rrset = iter->rrset->next;
+			} else {
+				iter->domain = domain_next(iter->domain);
+				if(iter->domain != NULL &&
+				   dname_is_subdomain(domain_dname(iter->domain),
+				                      domain_dname(iter->zone->apex)))
+				{
+					iter->rrset = iter->domain->rrsets;
+				}
+			}
+		}
+	}
+
+	assert(iter->rrset == NULL);
+	assert(iter->domain == NULL);
+	iter->index = -1;
+
+	return NULL;
+}
