@@ -16,7 +16,6 @@
 #include "namedb.h"
 #include "nsd.h"
 #include "answer.h"
-#include "udbzone.h"
 #include "options.h"
 
 #define NSEC3_RDATA_BITMAP 5
@@ -27,8 +26,12 @@ cmp_hash_tree(const void* x, const void* y)
 {
 	const domain_type* a = (const domain_type*)x;
 	const domain_type* b = (const domain_type*)y;
-	return memcmp(a->nsec3->nsec3_hash, b->nsec3->nsec3_hash,
-		NSEC3_HASH_LEN);
+	if(!a->nsec3) return (b->nsec3?-1:0);
+	if(!b->nsec3) return 1;
+	if(!a->nsec3->hash_wc) return (b->nsec3->hash_wc?-1:0);
+	if(!b->nsec3->hash_wc) return 1;
+	return memcmp(a->nsec3->hash_wc->hash.hash,
+		b->nsec3->hash_wc->hash.hash, NSEC3_HASH_LEN);
 }
 
 /* compare nsec3 hashes in nsec3 wc tree */
@@ -37,8 +40,12 @@ cmp_wchash_tree(const void* x, const void* y)
 {
 	const domain_type* a = (const domain_type*)x;
 	const domain_type* b = (const domain_type*)y;
-	return memcmp(a->nsec3->nsec3_wc_hash, b->nsec3->nsec3_wc_hash,
-		NSEC3_HASH_LEN);
+	if(!a->nsec3) return (b->nsec3?-1:0);
+	if(!b->nsec3) return 1;
+	if(!a->nsec3->hash_wc) return (b->nsec3->hash_wc?-1:0);
+	if(!b->nsec3->hash_wc) return 1;
+	return memcmp(a->nsec3->hash_wc->wc.hash,
+		b->nsec3->hash_wc->wc.hash, NSEC3_HASH_LEN);
 }
 
 /* compare nsec3 hashes in nsec3 ds tree */
@@ -47,8 +54,12 @@ cmp_dshash_tree(const void* x, const void* y)
 {
 	const domain_type* a = (const domain_type*)x;
 	const domain_type* b = (const domain_type*)y;
-	return memcmp(a->nsec3->nsec3_ds_parent_hash,
-		b->nsec3->nsec3_ds_parent_hash, NSEC3_HASH_LEN);
+	if(!a->nsec3) return (b->nsec3?-1:0);
+	if(!b->nsec3) return 1;
+	if(!a->nsec3->ds_parent_hash) return (b->nsec3->ds_parent_hash?-1:0);
+	if(!b->nsec3->ds_parent_hash) return 1;
+	return memcmp(a->nsec3->ds_parent_hash->hash,
+		b->nsec3->ds_parent_hash->hash, NSEC3_HASH_LEN);
 }
 
 /* compare base32-encoded nsec3 hashes in nsec3 rr tree, they are
@@ -59,9 +70,9 @@ cmp_nsec3_tree(const void* x, const void* y)
 	const domain_type* a = (const domain_type*)x;
 	const domain_type* b = (const domain_type*)y;
 	/* labelcount + 32long label */
-	assert(dname_name(a->dname)[0] == 32);
-	assert(dname_name(b->dname)[0] == 32);
-	return memcmp(dname_name(a->dname), dname_name(b->dname), 33);
+	assert(dname_name(domain_dname_const(a))[0] == 32);
+	assert(dname_name(domain_dname_const(b))[0] == 32);
+	return memcmp(dname_name(domain_dname_const(a)), dname_name(domain_dname_const(b)), 33);
 }
 
 void nsec3_zone_trees_create(struct region* region, zone_type* zone)
@@ -76,23 +87,19 @@ void nsec3_zone_trees_create(struct region* region, zone_type* zone)
 		zone->dshashtree = rbtree_create(region, cmp_dshash_tree);
 }
 
-void nsec3_hash_tree_clear(struct zone* zone)
-{
-	hash_tree_clear(zone->nsec3tree);
-	hash_tree_clear(zone->hashtree);
-	hash_tree_clear(zone->wchashtree);
-	hash_tree_clear(zone->dshashtree);
-}
-
 static void
 detect_nsec3_params(rr_type* nsec3_apex,
 	const unsigned char** salt, int* salt_len, int* iter)
 {
 	assert(salt && salt_len && iter);
 	assert(nsec3_apex);
-	*salt_len = rdata_atom_data(nsec3_apex->rdatas[3])[0];
-	*salt = (unsigned char*)(rdata_atom_data(nsec3_apex->rdatas[3])+1);
-	*iter = read_uint16(rdata_atom_data(nsec3_apex->rdatas[2]));
+	if(nsec3_apex->rdlength < 5)
+		return;
+	if(nsec3_apex->rdlength < 5 + (uint16_t)(nsec3_apex->rdata[4]))
+		return;
+	*salt_len = nsec3_apex->rdata[4];
+	*salt = (unsigned char*)(nsec3_apex->rdata + 5);
+	*iter = read_uint16(nsec3_apex->rdata + 2);
 }
 
 const dname_type *
@@ -115,6 +122,7 @@ nsec3_hash_and_store(zone_type* zone, const dname_type* dname, uint8_t* store)
 
 	detect_nsec3_params(zone->nsec3_param, &nsec3_salt,
 		&nsec3_saltlength, &nsec3_iterations);
+	assert(nsec3_iterations >= 0 && nsec3_iterations <= 65536);
 	iterated_hash((unsigned char*)store, nsec3_salt, nsec3_saltlength,
 		dname_name(dname), dname->name_size, nsec3_iterations);
 }
@@ -123,49 +131,65 @@ nsec3_hash_and_store(zone_type* zone, const dname_type* dname, uint8_t* store)
 
 /** find hash or create it and store it */
 static void
-nsec3_lookup_hash_and_wc(zone_type* zone, const dname_type* dname,
-	domain_type* domain, region_type* tmpregion)
+nsec3_lookup_hash_and_wc(region_type* region, zone_type* zone,
+	const dname_type* dname, domain_type* domain, region_type* tmpregion)
 {
 	const dname_type* wcard;
-	if(domain->nsec3->have_nsec3_hash && domain->nsec3->have_nsec3_wc_hash) {
+	if(domain->nsec3->hash_wc) {
 		return;
 	}
 	/* lookup failed; disk failure or so */
-	nsec3_hash_and_store(zone, dname, domain->nsec3->nsec3_hash);
-	domain->nsec3->have_nsec3_hash = 1;
+	domain->nsec3->hash_wc = (nsec3_hash_wc_node_type *)
+		region_alloc(region, sizeof(nsec3_hash_wc_node_type));
+	domain->nsec3->hash_wc->hash.node.key = NULL;
+	domain->nsec3->hash_wc->wc.node.key = NULL;
+	nsec3_hash_and_store(zone, dname, domain->nsec3->hash_wc->hash.hash);
 	wcard = dname_parse(tmpregion, "*");
 	wcard = dname_concatenate(tmpregion, wcard, dname);
-	nsec3_hash_and_store(zone, wcard, domain->nsec3->nsec3_wc_hash);
-	domain->nsec3->have_nsec3_wc_hash = 1;
+	nsec3_hash_and_store(zone, wcard, domain->nsec3->hash_wc->wc.hash);
 }
 
 static void
-nsec3_lookup_hash_ds(zone_type* zone, const dname_type* dname,
-	domain_type* domain)
+nsec3_lookup_hash_ds(region_type* region, zone_type* zone,
+	const dname_type* dname, domain_type* domain)
 {
-	if(domain->nsec3->have_nsec3_ds_parent_hash) {
+	if(domain->nsec3->ds_parent_hash) {
 		return;
 	}
 	/* lookup failed; disk failure or so */
-	nsec3_hash_and_store(zone, dname, domain->nsec3->nsec3_ds_parent_hash);
-	domain->nsec3->have_nsec3_ds_parent_hash = 1;
+	domain->nsec3->ds_parent_hash = (nsec3_hash_node_type *)
+		region_alloc(region, sizeof(nsec3_hash_node_type));
+	domain->nsec3->ds_parent_hash->node.key = NULL;
+	nsec3_hash_and_store(zone, dname, domain->nsec3->ds_parent_hash->hash);
 }
 
 static int
 nsec3_has_soa(rr_type* rr)
 {
-	if(rdata_atom_size(rr->rdatas[NSEC3_RDATA_BITMAP]) >= 3 && /* has types in bitmap */
-		rdata_atom_data(rr->rdatas[NSEC3_RDATA_BITMAP])[0] == 0 && /* first window = 0, */
-		/* [1]: bitmap length must be >= 1 */
-		/* [2]: bit[6] = SOA, thus mask first bitmap octet with 0x02 */
-		rdata_atom_data(rr->rdatas[NSEC3_RDATA_BITMAP])[2]&0x02) { /* SOA bit set */
+	size_t offset;
+	/* byte + byte + short + string + string + bitmap */
+	if(rr->rdlength < 6)
+		return 0;
+	offset = 4;
+	if(rr->rdlength < offset+1 ||
+		rr->rdlength < offset+1+rr->rdata[offset])
+		return 0;
+	offset += 1 + rr->rdata[offset];
+	if(rr->rdlength < offset+1 ||
+		rr->rdlength < offset+1+rr->rdata[offset])
+		return 0;
+	offset += 1 + rr->rdata[offset];
+	if (rr->rdlength >= offset + 3 && /* has types in bitmap */
+			rr->rdata[offset] == 0 && /* first window = 0, */
+	    /* [1]: bitmap length must be >= 1 */
+	    /* [2]: bit[6] = SOA, thus mask first bitmap octet with 0x02 */
+	    (rr->rdata[offset+2] & 0x02))
 		return 1;
-	}
 	return 0;
 }
 
 static rr_type*
-check_apex_soa(namedb_type* namedb, zone_type *zone)
+check_apex_soa(namedb_type* namedb, zone_type *zone, int nolog)
 {
 	uint8_t h[NSEC3_HASH_LEN];
 	domain_type* domain;
@@ -179,100 +203,131 @@ check_apex_soa(namedb_type* namedb, zone_type *zone)
 	hashed_apex = nsec3_b32_create(tmpregion, zone, h);
 	domain = domain_table_find(namedb->domains, hashed_apex);
 	if(!domain) {
-		log_msg(LOG_ERR, "%s NSEC3PARAM entry has no hash(apex).",
-			domain_to_string(zone->apex));
-		log_msg(LOG_ERR, "hash(apex)= %s",
-			dname_to_string(hashed_apex, NULL));
+		if(!nolog) {
+			log_msg(LOG_ERR, "%s NSEC3PARAM entry has no hash(apex).",
+				domain_to_string(zone->apex));
+			log_msg(LOG_ERR, "hash(apex)= %s",
+				dname_to_string(hashed_apex, NULL));
+		}
 		region_destroy(tmpregion);
 		return NULL;
 	}
 	nsec3_rrset = domain_find_rrset(domain, zone, TYPE_NSEC3);
 	if(!nsec3_rrset) {
-		log_msg(LOG_ERR, "%s NSEC3PARAM entry: hash(apex) has no NSEC3 RRset.",
-			domain_to_string(zone->apex));
-		log_msg(LOG_ERR, "hash(apex)= %s",
-			dname_to_string(hashed_apex, NULL));
+		if(!nolog) {
+			log_msg(LOG_ERR, "%s NSEC3PARAM entry: hash(apex) has no NSEC3 RRset.",
+				domain_to_string(zone->apex));
+			log_msg(LOG_ERR, "hash(apex)= %s",
+				dname_to_string(hashed_apex, NULL));
+		}
 		region_destroy(tmpregion);
 		return NULL;
 	}
 	for(j=0; j<nsec3_rrset->rr_count; j++) {
-		if(nsec3_has_soa(&nsec3_rrset->rrs[j])) {
+		if(nsec3_has_soa(nsec3_rrset->rrs[j])) {
 			region_destroy(tmpregion);
-			return &nsec3_rrset->rrs[j];
+			return nsec3_rrset->rrs[j];
 		}
 	}
-	log_msg(LOG_ERR, "%s NSEC3PARAM entry: hash(apex) NSEC3 has no SOA flag.",
-		domain_to_string(zone->apex));
-	log_msg(LOG_ERR, "hash(apex)= %s",
-		dname_to_string(hashed_apex, NULL));
+	if(!nolog) {
+		log_msg(LOG_ERR, "%s NSEC3PARAM entry: hash(apex) NSEC3 has no SOA flag.",
+			domain_to_string(zone->apex));
+		log_msg(LOG_ERR, "hash(apex)= %s",
+			dname_to_string(hashed_apex, NULL));
+	}
 	region_destroy(tmpregion);
 	return NULL;
 }
 
+static void
+nsec3param_to_str(struct rr* rr, char* str, size_t buflen)
+{
+	size_t len;
+	if(rr->rdlength < 5 || rr->rdlength < 5 + (uint16_t)rr->rdata[4]) {
+		str[0]=0;
+		return;
+	}
+	len = snprintf(str, buflen, "%u %u %u ",
+		(unsigned)rr->rdata[0], (unsigned)rr->rdata[1],
+		(unsigned)read_uint16(rr->rdata+2));
+	if(rr->rdata[4] == 0) {
+		if(buflen > len + 2)
+			str[len++] = '-';
+	} else {
+		len += hex_ntop(rr->rdata+5, rr->rdata[4], str+len, buflen-len-1);
+	}
+	if(buflen > len + 1) {
+		str[len] = 0;
+	} else {
+		assert(buflen > 0);
+		str[buflen-1] = 0;
+	}
+}
+
 static struct rr*
-udb_zone_find_nsec3param(udb_base* udb, udb_ptr* uz, struct zone* z)
+db_find_nsec3param(struct namedb* db, struct zone* z, struct rr* avoid_rr,
+	int checkchain)
 {
-	udb_ptr urr;
 	unsigned i;
 	rrset_type* rrset = domain_find_rrset(z->apex, z, TYPE_NSEC3PARAM);
 	if(!rrset) /* no NSEC3PARAM in mem */
 		return NULL;
-	udb_ptr_new(&urr, udb, &ZONE(uz)->nsec3param);
-	if(!urr.data || RR(&urr)->len < 5) {
-		/* no NSEC3PARAM in udb */
-		udb_ptr_unlink(&urr, udb);
-		return NULL;
-	}
-	/* find matching NSEC3PARAM RR in memory */
+	/* find first nsec3param we can support (SHA1, no flags) */
 	for(i=0; i<rrset->rr_count; i++) {
-		/* if this RR matches the udb RR then we are done */
-		rdata_atom_type* rd = rrset->rrs[i].rdatas;
-		if(RR(&urr)->wire[0] == rdata_atom_data(rd[0])[0] && /*alg*/
-		   RR(&urr)->wire[1] == rdata_atom_data(rd[1])[0] && /*flg*/
-		   RR(&urr)->wire[2] == rdata_atom_data(rd[2])[0] && /*iter*/
-		   RR(&urr)->wire[3] == rdata_atom_data(rd[2])[1] &&
-		   RR(&urr)->wire[4] == rdata_atom_data(rd[3])[0] && /*slen*/
-		   RR(&urr)->len >= 5 + RR(&urr)->wire[4] &&
-		   memcmp(RR(&urr)->wire+5, rdata_atom_data(rd[3])+1,
-			rdata_atom_data(rd[3])[0]) == 0) {
-			udb_ptr_unlink(&urr, udb);
-			return &rrset->rrs[i];
+		/* do not use the RR that is going to be deleted (in IXFR) */
+		if(rrset->rrs[i] == avoid_rr) continue;
+		if(rrset->rrs[i]->rdlength < 5) continue; /* require salt field */
+		if(rrset->rrs[i]->rdata[0] == NSEC3_SHA1_HASH &&
+			rrset->rrs[i]->rdata[1] == 0) {
+			if(checkchain) {
+				z->nsec3_param = rrset->rrs[i];
+				if(!check_apex_soa(db, z, 1)) {
+					char str[MAX_RDLENGTH*2+16];
+					nsec3param_to_str(z->nsec3_param,
+						str, sizeof(str));
+					VERBOSITY(1, (LOG_WARNING, "zone %s NSEC3PARAM %s has broken chain, ignoring", domain_to_string(z->apex), str));
+					continue; /* don't use broken chain */
+				}
+			}
+			if(2 <= verbosity) {
+				char str[MAX_RDLENGTH*2+16];
+				nsec3param_to_str(rrset->rrs[i], str,
+					sizeof(str));
+				VERBOSITY(2, (LOG_INFO, "rehash of zone %s with parameters %s",
+					domain_to_string(z->apex), str));
+			}
+			return rrset->rrs[i];
 		}
 	}
-	udb_ptr_unlink(&urr, udb);
 	return NULL;
 }
 
 void
-nsec3_find_zone_param(struct namedb* db, struct zone* zone, udb_ptr* z)
+nsec3_find_zone_param(struct namedb* db, struct zone* zone,
+	struct rr* avoid_rr, int checkchain)
 {
-	/* get nsec3param RR from udb */
-	zone->nsec3_param = udb_zone_find_nsec3param(db->udb, z, zone);
+	/* avoid using the rr that is going to be deleted, avoid_rr */
+	zone->nsec3_param = db_find_nsec3param(db, zone, avoid_rr, checkchain);
 }
 
 /* check params ok for one RR */
 static int
-nsec3_rdata_params_ok(rdata_atom_type* prd, rdata_atom_type* rd)
+nsec3_rdata_params_ok(const rr_type *prr, const rr_type* rr)
 {
-	return (rdata_atom_data(rd[0])[0] ==
-		rdata_atom_data(prd[0])[0] && /* hash algo */
-	   rdata_atom_data(rd[2])[0] ==
-		rdata_atom_data(prd[2])[0] && /* iterations 0 */
-	   rdata_atom_data(rd[2])[1] ==
-		rdata_atom_data(prd[2])[1] && /* iterations 1 */
-	   rdata_atom_data(rd[3])[0] ==
-		rdata_atom_data(prd[3])[0] && /* salt length */
-	   memcmp(rdata_atom_data(rd[3])+1,
-		rdata_atom_data(prd[3])+1, rdata_atom_data(rd[3])[0])
-		== 0 );
+	if(prr->rdlength > rr->rdlength)
+		return 0; /* The salt has to fit */
+	if(prr->rdlength < 5)
+		return 0; /* Malformed */
+	return prr->rdata[0] == rr->rdata[0] &&
+		(memcmp(prr->rdata+2, rr->rdata+2, prr->rdlength-2) == 0);
 }
 
 int
 nsec3_rr_uses_params(rr_type* rr, zone_type* zone)
 {
-	if(!rr || rr->rdata_count < 4)
+	if(!rr || rr->rdlength < 6)
 		return 0;
-	return nsec3_rdata_params_ok(zone->nsec3_param->rdatas, rr->rdatas);
+	return nsec3_rdata_params_ok(zone->nsec3_param, rr);
 }
 
 int
@@ -284,7 +339,7 @@ nsec3_in_chain_count(domain_type* domain, zone_type* zone)
 	if(!rrset || !zone->nsec3_param)
 		return 0; /* no NSEC3s, none in the chain */
 	for(i=0; i<rrset->rr_count; i++) {
-		if(nsec3_rr_uses_params(&rrset->rrs[i], zone))
+		if(nsec3_rr_uses_params(rrset->rrs[i], zone))
 			count++;
 	}
 	return count;
@@ -295,18 +350,43 @@ nsec3_chain_find_prev(struct zone* zone, struct domain* domain)
 {
 	if(domain->nsec3 && domain->nsec3->nsec3_node.key) {
 		/* see if there is a prev */
-		rbnode_t* r = rbtree_previous(&domain->nsec3->nsec3_node);
+		rbnode_type* r = rbtree_previous(&domain->nsec3->nsec3_node);
 		if(r != RBTREE_NULL) {
 			/* found a previous, which is not the root-node in
 			 * the prehash tree (and thus points to the tree) */
 			return (domain_type*)r->key;
 		}
 	}
-	if(zone->nsec3_last)
+	if(zone->nsec3_last && zone->nsec3_last != domain)
 		return zone->nsec3_last;
 	return NULL;
 }
 
+
+/** clear hash tree. Called from nsec3_clear_precompile() only. */
+static void
+hash_tree_clear(rbtree_type* tree)
+{
+	if(!tree) return;
+
+	/* Previously (before commit 4ca61188b3f7a0e077476875810d18a5d439871f
+	 * and/or svn commit 4776) prehashes and corresponding rbtree nodes
+	 * were part of struct nsec3_domain_data. Clearing the hash_tree would
+	 * then mean setting the key value of the nodes to NULL to indicate
+	 * absence of the prehash.
+	 * But since prehash structs are separatly allocated, this is no longer
+	 * necessary as currently the prehash structs are simply recycled and 
+	 * NULLed.
+	 *
+	 * rbnode_type* n;
+	 * for(n=rbtree_first(tree); n!=RBTREE_NULL; n=rbtree_next(n)) {
+	 *	n->key = NULL;
+	 * }
+	 */
+	tree->count = 0;
+	tree->root = RBTREE_NULL;
+}
+
 void
 nsec3_clear_precompile(struct namedb* db, zone_type* zone)
 {
@@ -324,22 +404,27 @@ nsec3_clear_precompile(struct namedb* db, zone_type* zone)
 	walk = zone->apex;
 	while(walk && domain_is_subdomain(walk, zone->apex)) {
 		if(walk->nsec3) {
-			if(nsec3_domain_part_of_zone(walk, zone)) {
+			if(nsec3_condition_hash(walk, zone)) {
 				walk->nsec3->nsec3_node.key = NULL;
 				walk->nsec3->nsec3_cover = NULL;
 				walk->nsec3->nsec3_wcard_child_cover = NULL;
 				walk->nsec3->nsec3_is_exact = 0;
-				walk->nsec3->have_nsec3_hash = 0;
-				walk->nsec3->have_nsec3_wc_hash = 0;
-				walk->nsec3->hash_node.key = NULL;
-				walk->nsec3->wchash_node.key = NULL;
+				if (walk->nsec3->hash_wc) {
+					region_recycle(db->domains->region,
+						walk->nsec3->hash_wc,
+						sizeof(nsec3_hash_wc_node_type));
+					walk->nsec3->hash_wc = NULL;
+				}
 			}
-			if(!walk->parent ||
-				nsec3_domain_part_of_zone(walk->parent, zone)) {
+			if(nsec3_condition_dshash(walk, zone)) {
 				walk->nsec3->nsec3_ds_parent_cover = NULL;
 				walk->nsec3->nsec3_ds_parent_is_exact = 0;
-				walk->nsec3->have_nsec3_ds_parent_hash = 0;
-				walk->nsec3->dshash_node.key = NULL;
+				if (walk->nsec3->ds_parent_hash) {
+					region_recycle(db->domains->region,
+						walk->nsec3->ds_parent_hash,
+						sizeof(nsec3_hash_node_type));
+					walk->nsec3->ds_parent_hash = NULL;
+				}
 			}
 		}
 		walk = domain_next(walk);
@@ -373,7 +458,8 @@ nsec3_condition_dshash(domain_type* d, zone_type* z)
 {
 	return d->is_existing && !domain_has_only_NSEC3(d, z) &&
 		(domain_find_rrset(d, z, TYPE_DS) ||
-		domain_find_rrset(d, z, TYPE_NS)) && d != z->apex;
+		domain_find_rrset(d, z, TYPE_NS)) && d != z->apex
+		&& nsec3_domain_part_of_zone(d->parent, z);
 }
 
 zone_type*
@@ -393,7 +479,7 @@ nsec3_tree_zone(namedb_type* db, domain_type* d)
 					rrset_rrtype(rrset) == TYPE_DNSKEY ||
 					rrset_rrtype(rrset) == TYPE_NSEC3PARAM)
 					return rrset->zone;
-			return namedb_find_zone(db, d->dname);
+			return namedb_find_zone(db, domain_dname(d));
 		}
 		d = d->parent;
 	}
@@ -406,21 +492,40 @@ nsec3_tree_dszone(namedb_type* db, domain_type* d)
 	/* the DStree does not contain nodes with d==z->apex */
 	if(d->is_apex)
 		d = d->parent;
-	return nsec3_tree_zone(db, d);
+	while (d) {
+		if (d->is_apex) {
+			zone_type *zone = NULL;
+			for (rrset_type *rrset = d->rrsets; rrset; rrset = rrset->next)
+				if (rrset_rrtype(rrset) == TYPE_SOA ||
+				    rrset_rrtype(rrset) == TYPE_DNSKEY ||
+				    rrset_rrtype(rrset) == TYPE_NSEC3PARAM)
+					zone = rrset->zone;
+			if (!zone)
+				zone = namedb_find_zone(db, domain_dname(d));
+			if (zone && zone->dshashtree)
+				return zone;
+		}
+		d = d->parent;
+	}
+	return NULL;
 }
 
 int
 nsec3_find_cover(zone_type* zone, uint8_t* hash, size_t hashlen,
 	domain_type** result)
 {
-	rbnode_t* r = NULL;
+	rbnode_type* r = NULL;
 	int exact;
 	domain_type d;
 	uint8_t n[48];
 
 	/* nsec3tree is sorted by b32 encoded domain name of the NSEC3 */
 	b32_ntop(hash, hashlen, (char*)(n+5), sizeof(n)-5);
+#ifdef USE_RADIX_TREE
 	d.dname = (dname_type*)n;
+#else
+	d.node.key = n;
+#endif
 	n[0] = 34; /* name_size */
 	n[1] = 2; /* label_count */
 	n[2] = 0; /* label_offset[0] */
@@ -448,25 +553,26 @@ nsec3_precompile_domain(struct namedb* db, struct domain* domain,
 	allocate_domain_nsec3(db->domains, domain);
 
 	/* hash it */
-	nsec3_lookup_hash_and_wc(zone, domain_dname(domain), domain, tmpregion);
+	nsec3_lookup_hash_and_wc(db->region,
+		zone, domain_dname(domain), domain, tmpregion);
 
 	/* add into tree */
 	zone_add_domain_in_hash_tree(db->region, &zone->hashtree,
-		cmp_hash_tree, domain, &domain->nsec3->hash_node);
+		cmp_hash_tree, domain, &domain->nsec3->hash_wc->hash.node);
 	zone_add_domain_in_hash_tree(db->region, &zone->wchashtree,
-		cmp_wchash_tree, domain, &domain->nsec3->wchash_node);
+		cmp_wchash_tree, domain, &domain->nsec3->hash_wc->wc.node);
 
 	/* lookup in tree cover ptr (or exact) */
-	exact = nsec3_find_cover(zone, domain->nsec3->nsec3_hash,
-		sizeof(domain->nsec3->nsec3_hash), &result);
+	exact = nsec3_find_cover(zone, domain->nsec3->hash_wc->hash.hash,
+		sizeof(domain->nsec3->hash_wc->hash.hash), &result);
 	domain->nsec3->nsec3_cover = result;
 	if(exact)
 		domain->nsec3->nsec3_is_exact = 1;
 	else	domain->nsec3->nsec3_is_exact = 0;
 
 	/* find cover for *.domain for wildcard denial */
-	exact = nsec3_find_cover(zone, domain->nsec3->nsec3_wc_hash,
-		sizeof(domain->nsec3->nsec3_wc_hash), &result);
+	(void)nsec3_find_cover(zone, domain->nsec3->hash_wc->wc.hash,
+		sizeof(domain->nsec3->hash_wc->wc.hash), &result);
 	domain->nsec3->nsec3_wcard_child_cover = result;
 }
 
@@ -480,17 +586,15 @@ nsec3_precompile_domain_ds(struct namedb* db, struct domain* domain,
 
 	/* hash it : it could have different hash parameters then the
 	   other hash for this domain name */
-	nsec3_lookup_hash_ds(zone, domain_dname(domain), domain);
+	nsec3_lookup_hash_ds(db->region, zone, domain_dname(domain), domain);
 	/* lookup in tree cover ptr (or exact) */
-	exact = nsec3_find_cover(zone, domain->nsec3->nsec3_ds_parent_hash,
-		sizeof(domain->nsec3->nsec3_ds_parent_hash), &result);
-	if(exact)
-		domain->nsec3->nsec3_ds_parent_is_exact = 1;
-	else 	domain->nsec3->nsec3_ds_parent_is_exact = 0;
+	exact = nsec3_find_cover(zone, domain->nsec3->ds_parent_hash->hash,
+		sizeof(domain->nsec3->ds_parent_hash->hash), &result);
+	domain->nsec3->nsec3_ds_parent_is_exact = exact != 0;
 	domain->nsec3->nsec3_ds_parent_cover = result;
 	/* add into tree */
 	zone_add_domain_in_hash_tree(db->region, &zone->dshashtree,
-		cmp_dshash_tree, domain, &domain->nsec3->dshash_node);
+		cmp_dshash_tree, domain, &domain->nsec3->ds_parent_hash->node);
 }
 
 static void
@@ -529,7 +633,7 @@ nsec3_precompile_newparam(namedb_type* db, zone_type* zone)
 	region_type* tmpregion = region_create(xalloc, free);
 	domain_type* walk;
 	time_t s = time(NULL);
-	unsigned n = 0, c = 0;
+	unsigned long n = 0, c = 0;
 
 	/* add nsec3s of chain to nsec3tree */
 	for(walk=zone->apex; walk && domain_is_subdomain(walk, zone->apex);
@@ -551,7 +655,8 @@ nsec3_precompile_newparam(namedb_type* db, zone_type* zone)
 		if(++c % ZONEC_PCT_COUNT == 0 && time(NULL) > s + ZONEC_PCT_TIME) {
 			s = time(NULL);
 			VERBOSITY(1, (LOG_INFO, "nsec3 %s %d %%",
-				zone->opts->name, c*100/n));
+				zone->opts->name,
+				(n==0)?0:(int)(c*((unsigned long)100)/n)));
 		}
 	}
 	region_destroy(tmpregion);
@@ -560,51 +665,52 @@ nsec3_precompile_newparam(namedb_type* db, zone_type* zone)
 void
 prehash_zone_complete(struct namedb* db, struct zone* zone)
 {
-	udb_ptr udbz;
-
 	/* robust clear it */
 	nsec3_clear_precompile(db, zone);
 	/* find zone settings */
 
 	assert(db && zone);
-	if(!udb_zone_search(db->udb, &udbz, dname_name(domain_dname(
-		zone->apex)), domain_dname(zone->apex)->name_size)) {
-		udb_ptr_init(&udbz, db->udb); /* zero the ptr */
-	}
-	nsec3_find_zone_param(db, zone, &udbz);
-	if(!zone->nsec3_param || !check_apex_soa(db, zone)) {
+	nsec3_find_zone_param(db, zone, NULL, 1);
+	if(!zone->nsec3_param || !check_apex_soa(db, zone, 0)) {
 		zone->nsec3_param = NULL;
 		zone->nsec3_last = NULL;
-		udb_ptr_unlink(&udbz, db->udb);
 		return;
 	}
-	udb_ptr_unlink(&udbz, db->udb);
+	nsec3_zone_trees_create(db->region, zone);
 	nsec3_precompile_newparam(db, zone);
 }
 
 static void
 init_lookup_key_hash_tree(domain_type* d, uint8_t* hash)
-{ memcpy(d->nsec3->nsec3_hash, hash, NSEC3_HASH_LEN); }
+{ memcpy(d->nsec3->hash_wc->hash.hash, hash, NSEC3_HASH_LEN); }
 
 static void
 init_lookup_key_wc_tree(domain_type* d, uint8_t* hash)
-{ memcpy(d->nsec3->nsec3_wc_hash, hash, NSEC3_HASH_LEN); }
+{ memcpy(d->nsec3->hash_wc->wc.hash, hash, NSEC3_HASH_LEN); }
 
 static void
 init_lookup_key_ds_tree(domain_type* d, uint8_t* hash)
-{ memcpy(d->nsec3->nsec3_ds_parent_hash, hash, NSEC3_HASH_LEN); }
+{ memcpy(d->nsec3->ds_parent_hash->hash, hash, NSEC3_HASH_LEN); }
 
 /* find first in the tree and true if the first to process it */
 static int
-process_first(rbtree_t* tree, uint8_t* hash, rbnode_t** p,
+process_first(rbtree_type* tree, uint8_t* hash, rbnode_type** p,
 	void (*init)(domain_type*, uint8_t*))
 {
 	domain_type d;
 	struct nsec3_domain_data n;
+	nsec3_hash_wc_node_type hash_wc;
+	nsec3_hash_node_type ds_parent_hash;
+
 	if(!tree) {
 		*p = RBTREE_NULL;
 		return 0;
 	}
+	hash_wc.hash.node.key = NULL;
+	hash_wc.wc.node.key = NULL;
+	n.hash_wc = &hash_wc;
+	ds_parent_hash.node.key = NULL;
+	n.ds_parent_hash = &ds_parent_hash;
 	d.nsec3 = &n;
 	init(&d, hash);
 	if(rbtree_find_less_equal(tree, &d, p)) {
@@ -622,15 +728,23 @@ process_first(rbtree_t* tree, uint8_t* hash, rbnode_t** p,
 
 /* set end pointer if possible */
 static void
-process_end(rbtree_t* tree, uint8_t* hash, rbnode_t** p,
+process_end(rbtree_type* tree, uint8_t* hash, rbnode_type** p,
 	void (*init)(domain_type*, uint8_t*))
 {
 	domain_type d;
 	struct nsec3_domain_data n;
+	nsec3_hash_wc_node_type hash_wc;
+	nsec3_hash_node_type ds_parent_hash;
+
 	if(!tree) {
 		*p = RBTREE_NULL;
 		return;
 	}
+	hash_wc.hash.node.key = NULL;
+	hash_wc.wc.node.key = NULL;
+	n.hash_wc = &hash_wc;
+	ds_parent_hash.node.key = NULL;
+	n.ds_parent_hash = &ds_parent_hash;
 	d.nsec3 = &n;
 	init(&d, hash);
 	if(rbtree_find_less_equal(tree, &d, p)) {
@@ -656,8 +770,8 @@ process_range(zone_type* zone, domain_type* start,
 {
 	/* start NULL means from first in tree */
 	/* end NULL means to last in tree */
-	rbnode_t *p = RBTREE_NULL, *pwc = RBTREE_NULL, *pds = RBTREE_NULL;
-	rbnode_t *p_end = RBTREE_NULL, *pwc_end = RBTREE_NULL, *pds_end = RBTREE_NULL;
+	rbnode_type *p = RBTREE_NULL, *pwc = RBTREE_NULL, *pds = RBTREE_NULL;
+	rbnode_type *p_end = RBTREE_NULL, *pwc_end = RBTREE_NULL, *pds_end = RBTREE_NULL;
 	/* because the nodes are on the prehashlist, the domain->nsec3 is
 	 * already allocated, and we need not allocate it here */
 	/* set start */
@@ -723,7 +837,7 @@ process_prehash_domain(domain_type* domain, zone_type* zone)
 	 * and set precompile pointers to point to this domain (or is_exact),
 	 * the first domain can be is_exact. If it is the last NSEC3, also
 	 * process the initial part (before the first) */
-	rbnode_t* nx;
+	rbnode_type* nx;
 
 	/* this domain is part of the prehash list and therefore the
 	 * domain->nsec3 is allocated and need not be allocated here */
@@ -751,6 +865,13 @@ void prehash_zone(struct namedb* db, struct zone* zone)
 		prehash_clear(db->domains);
 		return;
 	}
+	if(!check_apex_soa(db, zone, 1)) {
+		/* the zone fails apex soa check, prehash complete may
+		 * detect other valid chains */
+		prehash_clear(db->domains);
+		prehash_zone_complete(db, zone);
+		return;
+	}
 	/* process prehash list */
 	for(d = db->domains->prehash_list; d; d = d->nsec3->prehash_next) {
 		process_prehash_domain(d, zone);
@@ -758,7 +879,7 @@ void prehash_zone(struct namedb* db, struct zone* zone)
 	/* clear prehash list */
 	prehash_clear(db->domains);
 
-	if(!check_apex_soa(db, zone)) {
+	if(!check_apex_soa(db, zone, 0)) {
 		zone->nsec3_param = NULL;
 		zone->nsec3_last = NULL;
 	}
@@ -794,10 +915,39 @@ nsec3_add_nonexist_proof(struct query* query, struct answer* answer,
 	if(nsec3_find_cover(query->zone, hash, sizeof(hash), &cover))
 	{
 		/* exact match, hash collision */
+		domain_type* walk;
+		char hashbuf[512];
+		char reversebuf[512];
+		(void)b32_ntop(hash, sizeof(hash), hashbuf, sizeof(hashbuf));
+		snprintf(reversebuf, sizeof(reversebuf), "(no name in the zone hashes to this nsec3 record)");
+		walk = query->zone->apex;
+		while(walk) {
+			if(walk->nsec3 && walk->nsec3->nsec3_cover == cover) {
+				snprintf(reversebuf, sizeof(reversebuf),
+					"%s %s", domain_to_string(walk),
+					walk->nsec3->nsec3_is_exact?"exact":"no_exact_hash_match");
+				if(walk->nsec3->nsec3_is_exact)
+					break;
+			}
+			if(walk->nsec3 && walk->nsec3->nsec3_ds_parent_cover == cover) {
+				snprintf(reversebuf, sizeof(reversebuf),
+					"%s %s", domain_to_string(walk),
+					walk->nsec3->nsec3_ds_parent_is_exact?"exact":"no_exact_hash_match");
+				if(walk->nsec3->nsec3_ds_parent_is_exact)
+					break;
+			}
+			walk = domain_next(walk);
+		}
+
+
 		/* the hashed name of the query corresponds to an existing name. */
-		log_msg(LOG_ERR, "nsec3 hash collision for name=%s",
-			dname_to_string(to_prove, NULL));
+		VERBOSITY(3, (LOG_ERR, "nsec3 hash collision for name=%s hash=%s reverse=%s",
+			dname_to_string(to_prove, NULL), hashbuf, reversebuf));
 		RCODE_SET(query->packet, RCODE_SERVFAIL);
+		/* RFC 8914 - Extended DNS Errors
+		 * 4.21. Extended DNS Error Code 0 - Other */
+		ASSIGN_EDE_CODE_AND_STRING_LITERAL(query->edns.ede,
+			EDE_OTHER, "NSEC3 hash collision");
 		return;
 	}
 	else
@@ -843,7 +993,9 @@ nsec3_add_ds_proof(struct query *query, struct answer *answer,
 		/* use NSEC3 record from above the zone cut. */
 		nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
 			domain->nsec3->nsec3_ds_parent_cover);
-	} else if (!delegpt && domain->nsec3 && domain->nsec3->nsec3_is_exact) {
+	} else if (!delegpt && domain->nsec3 && domain->nsec3->nsec3_is_exact
+		&& nsec3_domain_part_of_zone(domain->nsec3->nsec3_cover,
+		query->zone)) {
 		nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
 			domain->nsec3->nsec3_cover);
 	} else {
@@ -868,12 +1020,21 @@ nsec3_add_ds_proof(struct query *query, struct answer *answer,
 				!prev_par->nsec3->nsec3_is_exact);
 			nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
 				prev_par->nsec3->nsec3_cover);
+		} else {
+			/* the exact case was handled earlier, so this is
+			 * with a closest-encloser proof, if in the part
+			 * before the else the closest encloser proof is done,
+			 * then we do not need to add a DS here because
+			 * the optout proof is already complete. If not,
+			 * we add the nsec3 here to complete the closest
+			 * encloser proof with a next closer */
+			/* add optout range from parent zone */
+			/* note: no check of optout bit, resolver checks it */
+			if(domain->nsec3) {
+				nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
+					domain->nsec3->nsec3_ds_parent_cover);
+			}
 		}
-		/* add optout range from parent zone */
-		/* note: no check of optout bit, resolver checks it */
-		if(domain->nsec3)
-			nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
-				domain->nsec3->nsec3_ds_parent_cover);
 	}
 }
 
@@ -896,6 +1057,17 @@ nsec3_answer_nodata(struct query* query, struct answer* answer,
 		}
 		/* query->zone must be the parent zone */
 		nsec3_add_ds_proof(query, answer, original, 0);
+		/* if the DS is from a wildcard match */
+		if (original==original->wildcard_child_closest_match
+			&& label_is_wildcard(dname_name(domain_dname(original)))) {
+			/* denial for wildcard is already there */
+			/* add parent proof to have a closest encloser proof for wildcard parent */
+			/* in other words: nsec3 matching closest encloser */
+			if(original->parent && original->parent->nsec3 &&
+				original->parent->nsec3->nsec3_is_exact)
+				nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
+					original->parent->nsec3->nsec3_cover);
+		}
 	}
 	/* the nodata is result from a wildcard match */
 	else if (original==original->wildcard_child_closest_match
@@ -915,9 +1087,20 @@ nsec3_answer_nodata(struct query* query, struct answer* answer,
 				original->nsec3->nsec3_cover);
 	}
 	else {	/* add nsec3 to prove rrset does not exist */
-		if(original->nsec3 && original->nsec3->nsec3_is_exact) {
+		if(original->nsec3) {
+			if(!original->nsec3->nsec3_is_exact) {
+				/* go up to an existing parent */
+				while(original->parent && original->parent->nsec3 && !original->parent->nsec3->nsec3_is_exact)
+					original = original->parent;
+			}
 			nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
 				original->nsec3->nsec3_cover);
+			if(!original->nsec3->nsec3_is_exact) {
+				if(original->parent && original->parent->nsec3 && original->parent->nsec3->nsec3_is_exact)
+				    nsec3_add_rrset(query, answer, AUTHORITY_SECTION,
+					original->parent->nsec3->nsec3_cover);
+
+			}
 		}
 	}
 }
@@ -940,9 +1123,9 @@ domain_has_only_NSEC3(struct domain* domain, struct zone* zone)
 	{
 		if(!zone || rrset->zone == zone)
 		{
-			if(rrset->rrs[0].type == TYPE_NSEC3)
+			if(rrset->rrs[0]->type == TYPE_NSEC3)
 				nsec3_seen = 1;
-			else if(rrset->rrs[0].type != TYPE_RRSIG)
+			else if(rrset->rrs[0]->type != TYPE_RRSIG)
 				return 0;
 		}
 		rrset = rrset->next;
@@ -984,6 +1167,10 @@ nsec3_answer_authoritative(struct domain** match, struct query *query,
 			/* wildcard exists below the domain */
 			/* wildcard and nsec3 domain clash. server failure. */
 			RCODE_SET(query->packet, RCODE_SERVFAIL);
+			/* RFC 8914 - Extended DNS Errors
+			 * 4.21. Extended DNS Error Code 0 - Other */
+			ASSIGN_EDE_CODE_AND_STRING_LITERAL(query->edns.ede,
+				EDE_OTHER, "Wildcard and NSEC3 domain clash");
 		}
 		return;
 	}
@@ -991,7 +1178,8 @@ nsec3_answer_authoritative(struct domain** match, struct query *query,
 #if 0
 		query->qtype != TYPE_NSEC3 &&
 #endif
-		domain_has_only_NSEC3(*match, query->zone))
+		(domain_has_only_NSEC3(*match, query->zone) ||
+		!domain_find_any_rrset(*match, query->zone)))
 	{
 		/* this looks like a NSEC3 domain, but is actually an empty non-terminal. */
 		nsec3_answer_nodata(query, answer, *match);
