@@ -62,7 +62,8 @@ chunk_get_last(void* base, udb_void chunk, int exp)
 static void
 chunk_set_last(void* base, udb_void chunk, int exp, uint8_t value)
 {
-	*((uint8_t*)UDB_REL(base, chunk+(1<<exp)-1)) = value;
+	assert(exp >= 0 && exp <= 63);
+	*((uint8_t*)UDB_REL(base, chunk+((uint64_t)1<<exp)-1)) = value;
 }
 
 /** create udb_base from a file descriptor (must be at start of file) */
@@ -70,7 +71,7 @@ udb_base*
 udb_base_create_fd(const char* fname, int fd, udb_walk_relptr_func walkfunc,
 	void* arg)
 {
-	uint64_t m;
+	uint64_t m, fsz;
 	udb_glob_d g;
 	ssize_t r;
 	udb_base* udb = (udb_base*)xalloc_zero(sizeof(*udb));
@@ -91,7 +92,8 @@ udb_base_create_fd(const char* fname, int fd, udb_walk_relptr_func walkfunc,
 	udb->fd = fd;
 	udb->ram_size = 1024;
 	udb->ram_mask = (int)udb->ram_size - 1;
-	udb->ram_hash = (udb_ptr**)xalloc_zero(sizeof(udb_ptr*)*udb->ram_size);
+	udb->ram_hash = (udb_ptr**)xalloc_array_zero(sizeof(udb_ptr*),
+		udb->ram_size);
 	if(!udb->ram_hash) {
 		free(udb->fname);
 		free(udb);
@@ -134,26 +136,49 @@ udb_base_create_fd(const char* fname, int fd, udb_walk_relptr_func walkfunc,
 	if(g.hsize > UDB_HEADER_SIZE) {
 		log_msg(LOG_WARNING, "%s: header size too large %d", fname,
 			(int)g.hsize);
-		log_msg(LOG_WARNING, "attempting to continue...");
+		goto fail;
 	}
-	if(g.clean_close != 0) {
+	if(g.clean_close != 1) {
 		log_msg(LOG_WARNING, "%s: not cleanly closed %d", fname,
 			(int)g.clean_close);
-		log_msg(LOG_WARNING, "attempting to continue...");
+		goto fail;
 	}
-	/* TODO check if too large (>4g on 32bit); mmap-usage would fail */
-	
+	if(g.dirty_alloc != 0) {
+		log_msg(LOG_WARNING, "%s: not cleanly closed (alloc:%d)", fname,
+			(int)g.dirty_alloc);
+		goto fail;
+	}
+
+	/* check file size correctly written, for 4.0.2 nsd.db failure */
+	fsz = (uint64_t)lseek(fd, (off_t)0, SEEK_END);
+	(void)lseek(fd, (off_t)0, SEEK_SET);
+	if(g.fsize != fsz) {
+		log_msg(LOG_WARNING, "%s: file size %llu but mmap header "
+			"has size %llu", fname, (unsigned long long)fsz,
+			(unsigned long long)g.fsize);
+		goto fail;
+	}
+
 	/* mmap it */
 	if(g.fsize < UDB_HEADER_SIZE || g.fsize < g.hsize) {
 		log_msg(LOG_ERR, "%s: file too short", fname);
 		goto fail;
 	}
+	if(g.fsize > (uint64_t)400*1024*1024*1024*1024) /* 400 Tb */ {
+		log_msg(LOG_WARNING, "%s: file size too large %llu",
+			fname, (unsigned long long)g.fsize);
+		goto fail;
+	}
 	udb->base_size = (size_t)g.fsize;
+#ifdef HAVE_MMAP
 	/* note the size_t casts must be there for portability, on some
 	 * systems the layout of memory is otherwise broken. */
 	udb->base = mmap(NULL, (size_t)udb->base_size,
 		(int)PROT_READ|PROT_WRITE, (int)MAP_SHARED,
 		(int)udb->fd, (off_t)0);
+#else
+	udb->base = MAP_FAILED; errno = ENOSYS;
+#endif
 	if(udb->base == MAP_FAILED) {
 		udb->base = NULL;
 		log_msg(LOG_ERR, "mmap(size %u) error: %s",
@@ -167,12 +192,13 @@ udb_base_create_fd(const char* fname, int fd, udb_walk_relptr_func walkfunc,
 	}
 
 	/* init completion */
-	udb->glob_data = (udb_glob_d*)(udb->base+sizeof(uint64_t));
+	udb->glob_data = (udb_glob_d*)((char*)udb->base+sizeof(uint64_t));
 	r = 0;
+	/* cannot be dirty because that is goto fail above */
 	if(udb->glob_data->dirty_alloc != udb_dirty_clean)
 		r = 1;
 	udb->alloc = udb_alloc_create(udb, (udb_alloc_d*)(
-		(void*)udb->glob_data+sizeof(*udb->glob_data)));
+		(char*)udb->glob_data+sizeof(*udb->glob_data)));
 	if(!udb->alloc) {
 		log_msg(LOG_ERR, "out of memory");
 		udb_base_free(udb);
@@ -183,6 +209,7 @@ udb_base_create_fd(const char* fname, int fd, udb_walk_relptr_func walkfunc,
 		udb_alloc_compact(udb, udb->alloc);
 		udb_base_sync(udb, 1);
 	}
+	udb->glob_data->clean_close = 0;
 
 	return udb;
 }
@@ -239,6 +266,7 @@ udb_base* udb_base_create_new(const char* fname, udb_walk_relptr_func walkfunc,
 	m = UDB_MAGIC;
 	udb_glob_init_new(&g);
 	udb_alloc_init_new(&a);
+	g.clean_close = 1;
 
 	/* write new data to file (closes fd on error) */
 	if(!write_fdata(fname, fd, &m, sizeof(m)))
@@ -257,6 +285,13 @@ udb_base* udb_base_create_new(const char* fname, udb_walk_relptr_func walkfunc,
 		close(fd);
 		return NULL;
 	}
+	/* truncate to the right size */
+	if(ftruncate(fd, (off_t)g.fsize) < 0) {
+		log_msg(LOG_ERR, "%s: ftruncate(%d): %s", fname,
+			(int)g.fsize, strerror(errno));
+		close(fd);
+		return NULL;
+	}
 	return udb_base_create_fd(fname, fd, walkfunc, arg);
 }
 
@@ -268,7 +303,9 @@ udb_base_shrink(udb_base* udb, uint64_t nsize)
 	udb->glob_data->fsize = nsize;
 	/* sync, does not *seem* to be required on Linux, but it is
 	   certainly required on OpenBSD.  Otherwise changed data is lost. */
+#ifdef HAVE_MMAP
 	msync(udb->base, udb->base_size, MS_ASYNC);
+#endif
 	if(ftruncate(udb->fd, (off_t)nsize) != 0) {
 		log_msg(LOG_ERR, "%s: ftruncate(%u) %s", udb->fname,
 			(unsigned)nsize, strerror(errno));
@@ -286,13 +323,16 @@ void udb_base_close(udb_base* udb)
 			udb_base_shrink(udb, nsize);
 	}
 	if(udb->fd != -1) {
+		udb->glob_data->clean_close = 1;
 		close(udb->fd);
 		udb->fd = -1;
 	}
 	if(udb->base) {
+#ifdef HAVE_MMAP
 		if(munmap(udb->base, udb->base_size) == -1) {
 			log_msg(LOG_ERR, "munmap: %s", strerror(errno));
 		}
+#endif
 		udb->base = NULL;
 	}
 }
@@ -324,10 +364,15 @@ void udb_base_free_keep_mmap(udb_base* udb)
 
 void udb_base_sync(udb_base* udb, int wait)
 {
+	if(!udb) return;
+#ifdef HAVE_MMAP
 	if(msync(udb->base, udb->base_size, wait?MS_SYNC:MS_ASYNC) != 0) {
 		log_msg(LOG_ERR, "msync(%s) error %s",
 			udb->fname, strerror(errno));
 	}
+#else
+	(void)wait;
+#endif
 }
 
 /** hash a chunk pointer */
@@ -383,20 +428,22 @@ grow_ram_hash(udb_base* udb, udb_ptr** newhash)
 
 void udb_base_link_ptr(udb_base* udb, udb_ptr* ptr)
 {
-	uint32_t i = chunk_hash_ptr(ptr->data) & udb->ram_mask;
-	assert((size_t)i < udb->ram_size);
+	uint32_t i;
 #ifdef UDB_CHECK
 	assert(udb_valid_dataptr(udb, ptr->data)); /* must be to whole chunk*/
 #endif
 	udb->ram_num++;
-	if(udb->ram_num == udb->ram_size && udb->ram_size<(size_t)0xefffffff) {
+	if(udb->ram_num == udb->ram_size && udb->ram_size<(size_t)0x7fffffff) {
 		/* grow the array, if allocation succeeds */
-		udb_ptr** newram = (udb_ptr**)xalloc_zero(sizeof(udb_ptr*)*
-			udb->ram_size*2);
+		udb_ptr** newram = (udb_ptr**)xalloc_array_zero(
+			sizeof(udb_ptr*), udb->ram_size*2);
 		if(newram) {
 			grow_ram_hash(udb, newram);
 		}
 	}
+	i = chunk_hash_ptr(ptr->data) & udb->ram_mask;
+	assert((size_t)i < udb->ram_size);
+
 	ptr->prev = NULL;
 	ptr->next = udb->ram_hash[i];
 	udb->ram_hash[i] = ptr;
@@ -469,6 +516,7 @@ uint8_t udb_base_get_userflags(udb_base* udb)
 static void*
 udb_base_remap(udb_base* udb, udb_alloc* alloc, uint64_t nsize)
 {
+#ifdef HAVE_MMAP
 	void* nb;
 	/* for use with valgrind, do not use mremap, but the other version */
 #ifdef MREMAP_MAYMOVE
@@ -507,14 +555,18 @@ udb_base_remap(udb_base* udb, udb_alloc* alloc, uint64_t nsize)
 		/* fix up realpointers in udb and alloc */
 		/* but mremap may have been nice and not move the base */
 		udb->base = nb;
-		udb->glob_data = (udb_glob_d*)(nb+sizeof(uint64_t));
+		udb->glob_data = (udb_glob_d*)((char*)nb+sizeof(uint64_t));
 		/* use passed alloc pointer because the udb->alloc may not
 		 * be initialized yet */
-		alloc->disk = (udb_alloc_d*)((void*)udb->glob_data
+		alloc->disk = (udb_alloc_d*)((char*)udb->glob_data
 			+sizeof(*udb->glob_data));
 	}
 	udb->base_size = nsize;
 	return nb;
+#else /* HAVE_MMAP */
+	(void)udb; (void)alloc; (void)nsize;
+	return NULL;
+#endif /* HAVE_MMAP */
 }
 
 void
@@ -576,8 +628,9 @@ int udb_exp_size(uint64_t a)
 		i >>= 1;
 		x ++;
 	}
+	assert( x>=0 && x<=63);
 	assert( ((uint64_t)1<<x) >= a);
-	assert( x==0 || ((uint64_t)1<<(x-1)) < a);
+	assert( x==0 || /* <<x-1 without negative number analyzer complaints: */ (((uint64_t)1<<x)>>1) < a);
 	return x;
 }
 
@@ -745,7 +798,7 @@ regen_ptrlist(void* base, udb_base* udb, udb_alloc* alloc,
 		if(exp == UDB_EXP_XL) {
 			assert(at != rb_old); /* should have been freed */
 			regen_its_ptrs(base, udb, atp,
-				((void*)atp)+sizeof(udb_xl_chunk_d),
+				((char*)atp)+sizeof(udb_xl_chunk_d),
 				sz-sizeof(udb_xl_chunk_d) - sizeof(uint64_t)*2,
 				rb_old, rb_new);
 			at += sz;
@@ -754,7 +807,7 @@ regen_ptrlist(void* base, udb_base* udb, udb_alloc* alloc,
 		} else { /* data chunk */
 			assert(at != rb_old); /* should have been freed */
 			regen_its_ptrs(base, udb, atp,
-				((void*)atp)+sizeof(udb_chunk_d),
+				((char*)atp)+sizeof(udb_chunk_d),
 				sz-sizeof(udb_chunk_d)-1, rb_old, rb_new);
 			at += sz;
 		}
@@ -1096,7 +1149,7 @@ grow_extra_check(udb_alloc* alloc, uint64_t ge)
 	return ge;
 }
 
-/** see if free space is enogh to warrant shrink (while file is open) */
+/** see if free space is enough to warrant shrink (while file is open) */
 static int
 enough_free(udb_alloc* alloc)
 {
@@ -1227,12 +1280,15 @@ static int
 udb_alloc_exp_needed(size_t sz)
 {
 	uint64_t asz = sz + sizeof(udb_chunk_d) + 1;
+	int exp;
 	if(asz > UDB_ALLOC_CHUNK_SIZE) {
 		return UDB_EXP_XL;
 	} else if(asz <= UDB_ALLOC_CHUNK_MINSIZE) {
 		return UDB_ALLOC_CHUNK_MINEXP;
 	}
-	return udb_exp_size(asz);
+	exp = udb_exp_size(asz);
+	assert(exp <= UDB_ALLOC_CHUNKS_MAX);
+	return exp;
 }
 
 udb_void udb_alloc_space(udb_alloc* alloc, size_t sz)
@@ -1261,7 +1317,7 @@ udb_void udb_alloc_space(udb_alloc* alloc, size_t sz)
 		return ret + sizeof(udb_chunk_d); /* ptr to data */
 	}
 	/* see if we can subdivide a larger chunk */
-	for(e2 = exp+1; e2 < UDB_ALLOC_CHUNKS_MAX; e2++)
+	for(e2 = exp+1; e2 <= UDB_ALLOC_CHUNKS_MAX; e2++)
 		if(alloc->disk->free[e2-UDB_ALLOC_CHUNK_MINEXP]) {
 			udb_void big, ret; /* udb_chunk_d */
 			alloc->udb->glob_data->dirty_alloc = udb_dirty_fl;
@@ -1296,7 +1352,7 @@ have_free_for(udb_alloc* alloc, int exp)
 	int e2;
 	if(alloc->disk->free[exp-UDB_ALLOC_CHUNK_MINEXP])
 		return exp;
-	for(e2 = exp+1; e2 < UDB_ALLOC_CHUNKS_MAX; e2++)
+	for(e2 = exp+1; e2 <= UDB_ALLOC_CHUNKS_MAX; e2++)
 		if(alloc->disk->free[e2-UDB_ALLOC_CHUNK_MINEXP]) {
 			return e2;
 		}
@@ -1501,7 +1557,7 @@ coagulate_and_push(void* base, udb_alloc* alloc, udb_void last, int exp,
 }
 
 /** attempt to compact the data and move free space to the end */
-static int
+int
 udb_alloc_compact(void* base, udb_alloc* alloc)
 {
 	udb_void last;
@@ -1510,6 +1566,9 @@ udb_alloc_compact(void* base, udb_alloc* alloc)
 	uint64_t at = alloc->disk->nextgrow;
 	udb_void xl_start = 0;
 	uint64_t xl_sz = 0;
+	if(alloc->udb->inhibit_compact)
+		return 1;
+	alloc->udb->useful_compact = 0;
 	while(at > alloc->udb->glob_data->hsize) {
 		/* grab last entry */
 		exp = (int)*((uint8_t*)UDB_REL(base, at-1));
@@ -1627,6 +1686,21 @@ udb_alloc_compact(void* base, udb_alloc* alloc)
 	return 1;
 }
 
+int
+udb_compact(udb_base* udb)
+{
+	if(!udb) return 1;
+	if(!udb->useful_compact) return 1;
+	DEBUG(DEBUG_DBACCESS, 1, (LOG_INFO, "Compacting database..."));
+	return udb_alloc_compact(udb->base, udb->alloc);
+}
+
+void udb_compact_inhibited(udb_base* udb, int inhibit)
+{
+	if(!udb) return;
+	udb->inhibit_compact = inhibit;
+}
+
 #ifdef UDB_CHECK
 /** check that rptrs are really zero before free */
 void udb_check_rptr_zero(void* base, udb_rel_ptr* p, void* arg)
@@ -1703,6 +1777,10 @@ int udb_alloc_free(udb_alloc* alloc, udb_void r, size_t sz)
 	if(fp->exp == UDB_EXP_XL) {
 		udb_free_xl(base, alloc, f, (udb_xl_chunk_d*)fp, sz);
 		/* compact */
+		if(alloc->udb->inhibit_compact) {
+			alloc->udb->useful_compact = 1;
+			return 1;
+		}
 		return udb_alloc_compact(base, alloc);
 	}
 	/* it is a regular chunk of 2**exp size */
@@ -1746,6 +1824,10 @@ int udb_alloc_free(udb_alloc* alloc, udb_void r, size_t sz)
 	}
 	alloc->udb->glob_data->dirty_alloc = udb_dirty_clean;
 	/* compact */
+	if(alloc->udb->inhibit_compact) {
+		alloc->udb->useful_compact = 1;
+		return 1;
+	}
 	return udb_alloc_compact(base, alloc);
 }
 
