@@ -12,14 +12,26 @@
 #include <assert.h>
 #include <ctype.h>
 #include <errno.h>
+#ifdef HAVE_OPENSSL_RAND_H
+#include <openssl/rand.h>
+#endif
 #include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#ifdef HAVE_SCHED_H
+#include <sched.h>
+#endif /* HAVE_SCHED_H */
+#ifdef HAVE_SYS_CPUSET_H
+#include <sys/cpuset.h>
+#endif /* HAVE_SYS_CPUSET_H */
 #ifdef HAVE_SYSLOG_H
 #include <syslog.h>
 #endif /* HAVE_SYSLOG_H */
 #include <unistd.h>
+#ifdef HAVE_SYS_RANDOM_H
+#include <sys/random.h>
+#endif
 
 #include "util.h"
 #include "region-allocator.h"
@@ -27,6 +39,8 @@
 #include "namedb.h"
 #include "rdata.h"
 #include "zonec.h"
+#include "nsd.h"
+#include "options.h"
 
 #ifdef USE_MMAP_ALLOC
 #include <sys/mman.h>
@@ -51,6 +65,16 @@ int verbosity = 0;
 static const char *global_ident = NULL;
 static log_function_type *current_log_function = log_file;
 static FILE *current_log_file = NULL;
+int log_time_asc = 1;
+int log_time_iso = 0;
+
+#ifdef USE_LOG_PROCESS_ROLE
+void
+log_set_process_role(const char *process_role)
+{
+	global_ident = process_role;
+}
+#endif
 
 void
 log_init(const char *ident)
@@ -135,7 +159,48 @@ log_file(int priority, const char *message)
 	}
 
 	/* Bug #104, add time_t timestamp */
-	fprintf(current_log_file, "[%d] %s[%d]: %s: %s",
+#if defined(HAVE_STRFTIME) && defined(HAVE_LOCALTIME_R)
+	if(log_time_asc) {
+		struct timeval tv;
+		char tmbuf[32];
+		tmbuf[0]=0;
+		tv.tv_usec = 0;
+		if(log_time_iso) {
+			char tzbuf[16];
+			tzbuf[0]=0;
+			/* log time in iso format */
+			if(gettimeofday(&tv, NULL) == 0) {
+				struct tm tm, *tm_p;
+				time_t now = (time_t)tv.tv_sec;
+				tm_p = localtime_r(&now, &tm);
+				strftime(tmbuf, sizeof(tmbuf), "%Y-%m-%dT%H:%M:%S",
+					tm_p);
+				if(strftime(tzbuf, sizeof(tzbuf), "%z", tm_p) == 5) {
+					/* put ':' in "+hh:mm" */
+					tzbuf[5] = tzbuf[4];
+					tzbuf[4] = tzbuf[3];
+					tzbuf[3] = ':';
+					tzbuf[6] = 0;
+				}
+			}
+			fprintf(current_log_file, "%s.%3.3d%s %s[%d]: %s: %s",
+				tmbuf, (int)tv.tv_usec/1000, tzbuf,
+				global_ident, (int) getpid(), priority_text, message);
+		} else {
+			/* log time in ascii format */
+			if(gettimeofday(&tv, NULL) == 0) {
+				struct tm tm;
+				time_t now = (time_t)tv.tv_sec;
+				strftime(tmbuf, sizeof(tmbuf), "%Y-%m-%d %H:%M:%S",
+					localtime_r(&now, &tm));
+			}
+			fprintf(current_log_file, "[%s.%3.3d] %s[%d]: %s: %s",
+				tmbuf, (int)tv.tv_usec/1000,
+				global_ident, (int) getpid(), priority_text, message);
+		}
+ 	} else
+#endif /* have time functions */
+		fprintf(current_log_file, "[%d] %s[%d]: %s: %s",
 		(int)time(NULL), global_ident, (int) getpid(), priority_text, message);
 	length = strlen(message);
 	if (length == 0 || message[length - 1] != '\n') {
@@ -153,6 +218,17 @@ log_syslog(int priority, const char *message)
 	log_file(priority, message);
 }
 
+void
+log_only_syslog(int priority, const char *message)
+{
+#ifdef HAVE_SYSLOG_H
+	syslog(priority, "%s", message);
+#else /* !HAVE_SYSLOG_H */
+	/* no syslog, use stderr */
+	log_file(priority, message);
+#endif
+}
+
 void
 log_set_log_function(log_function_type *log_function)
 {
@@ -197,7 +273,7 @@ clear_bit(uint8_t bits[], size_t index)
 }
 
 int
-get_bit(uint8_t bits[], size_t index)
+get_bit(const uint8_t bits[], size_t index)
 {
 	/*
 	 * The bits are counted from left to right, so bit #0 is the
@@ -228,6 +304,19 @@ lookup_by_id(lookup_table_type *table, int id)
 	return NULL;
 }
 
+char *
+xstrdup(const char *src)
+{
+	char *result = strdup(src);
+
+	if(!result) {
+		log_msg(LOG_ERR, "strdup failed: %s", strerror(errno));
+		exit(1);
+	}
+
+	return result;
+}
+
 void *
 xalloc(size_t size)
 {
@@ -240,6 +329,18 @@ xalloc(size_t size)
 	return result;
 }
 
+void *
+xmallocarray(size_t num, size_t size)
+{
+        void *result = reallocarray(NULL, num, size);
+
+        if (!result) {
+                log_msg(LOG_ERR, "reallocarray failed: %s", strerror(errno));
+                exit(1);
+        }
+        return result;
+}
+
 void *
 xalloc_zero(size_t size)
 {
@@ -251,6 +352,17 @@ xalloc_zero(size_t size)
 	return result;
 }
 
+void *
+xalloc_array_zero(size_t num, size_t size)
+{
+	void *result = calloc(num, size);
+	if (!result) {
+		log_msg(LOG_ERR, "calloc failed: %s", strerror(errno));
+		exit(1);
+	}
+	return result;
+}
+
 void *
 xrealloc(void *ptr, size_t size)
 {
@@ -352,6 +464,28 @@ write_socket(int s, const void *buf, size_t size)
 	return 1;
 }
 
+void get_time(struct timespec* t)
+{
+	struct timeval tv;
+#ifdef HAVE_CLOCK_GETTIME
+	/* first try nanosecond precision */
+	if(clock_gettime(CLOCK_REALTIME, t)>=0) {
+		return; /* success */
+	}
+	log_msg(LOG_ERR, "clock_gettime: %s", strerror(errno));
+#endif
+	/* try millisecond precision */
+	if(gettimeofday(&tv, NULL)>=0) {
+		t->tv_sec = tv.tv_sec;
+		t->tv_nsec = tv.tv_usec*1000;
+		return; /* success */
+	}
+	log_msg(LOG_ERR, "gettimeofday: %s", strerror(errno));
+	/* whole seconds precision */
+	t->tv_sec = time(0);
+	t->tv_nsec = 0;
+}
+
 int
 timespec_compare(const struct timespec *left,
 		 const struct timespec *right)
@@ -403,109 +537,6 @@ timespec_subtract(struct timespec *left,
 	}
 }
 
-uint32_t
-strtoserial(const char* nptr, const char** endptr)
-{
-	uint32_t i = 0;
-	uint32_t serial = 0;
-
-	for(*endptr = nptr; **endptr; (*endptr)++) {
-		switch (**endptr) {
-		case ' ':
-		case '\t':
-			break;
-		case '0':
-		case '1':
-		case '2':
-		case '3':
-		case '4':
-		case '5':
-		case '6':
-		case '7':
-		case '8':
-		case '9':
-			i *= 10;
-			i += (**endptr - '0');
-			break;
-		default:
-			break;
-		}
-	}
-	serial += i;
-	return serial;
-}
-
-uint32_t
-strtottl(const char *nptr, const char **endptr)
-{
-	uint32_t i = 0;
-	uint32_t seconds = 0;
-
-	for(*endptr = nptr; **endptr; (*endptr)++) {
-		switch (**endptr) {
-		case ' ':
-		case '\t':
-			break;
-		case 's':
-		case 'S':
-			seconds += i;
-			i = 0;
-			break;
-		case 'm':
-		case 'M':
-			seconds += i * 60;
-			i = 0;
-			break;
-		case 'h':
-		case 'H':
-			seconds += i * 60 * 60;
-			i = 0;
-			break;
-		case 'd':
-		case 'D':
-			seconds += i * 60 * 60 * 24;
-			i = 0;
-			break;
-		case 'w':
-		case 'W':
-			seconds += i * 60 * 60 * 24 * 7;
-			i = 0;
-			break;
-		case '0':
-		case '1':
-		case '2':
-		case '3':
-		case '4':
-		case '5':
-		case '6':
-		case '7':
-		case '8':
-		case '9':
-			i *= 10;
-			i += (**endptr - '0');
-			break;
-		default:
-			seconds += i;
-			/**
-			 * According to RFC2308, Section 8, the MSB
-			 * (sign bit) should be set to zero.
-			 * If we encounter a value larger than 2^31 -1,
-			 * we fall back to the default TTL.
-			 */
-			if ((seconds & MSB_32)) {
-				seconds = DEFAULT_TTL;
-			}
-			return seconds;
-		}
-	}
-	seconds += i;
-	if ((seconds & MSB_32)) {
-		seconds = DEFAULT_TTL;
-	}
-	return seconds;
-}
-
-
 ssize_t
 hex_ntop(uint8_t const *src, size_t srclength, char *target, size_t targsize)
 {
@@ -535,7 +566,8 @@ hex_pton(const char* src, uint8_t* target, size_t targsize)
 		return -1;
 	}
 	while(*src) {
-		if(!isxdigit((int)src[0]) || !isxdigit((int)src[1]))
+		if(!isxdigit((unsigned char)src[0]) ||
+			!isxdigit((unsigned char)src[1]))
 			return -1;
 		*t++ = hexdigit_to_int(src[0]) * 16 +
 			hexdigit_to_int(src[1]) ;
@@ -613,10 +645,10 @@ b32_ntop(uint8_t const *src, size_t srclength, char *target, size_t targsize)
 	}
 	if(srclength)
 	{
-		if(targsize < strlen(buf)+1)
+		size_t tlen = strlcpy(target, buf, targsize);
+		if (tlen >= targsize)
 			return -1;
-		strlcpy(target, buf, targsize);
-		len += strlen(buf);
+		len += tlen;
 	}
 	else if(targsize < 1)
 		return -1;
@@ -640,7 +672,7 @@ b32_pton(const char *src, uint8_t *target, size_t tsize)
 		if(p+5 >= tsize*8)
 		       return -1;
 
-		if(isspace(ch))
+		if(isspace((unsigned char)ch))
 			continue;
 
 		if(ch >= '0' && ch <= '9')
@@ -672,13 +704,13 @@ strip_string(char *str)
 	char *start = str;
 	char *end = str + strlen(str) - 1;
 
-	while (isspace(*start))
+	while (isspace((unsigned char)*start))
 		++start;
 	if (start > end) {
 		/* Completely blank. */
 		str[0] = '\0';
 	} else {
-		while (isspace(*end))
+		while (isspace((unsigned char)*end))
 			--end;
 		*++end = '\0';
 
@@ -712,58 +744,11 @@ hexdigit_to_int(char ch)
 	}
 }
 
-/* Number of days per month (except for February in leap years). */
-static const int mdays[] = {
-    31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31
-};
-
-static int
-is_leap_year(int year)
-{
-    return year % 4 == 0 && (year % 100 != 0 || year % 400 == 0);
-}
-
-static int
-leap_days(int y1, int y2)
-{
-    --y1;
-    --y2;
-    return (y2/4 - y1/4) - (y2/100 - y1/100) + (y2/400 - y1/400);
-}
-
-/*
- * Code adapted from Python 2.4.1 sources (Lib/calendar.py).
- */
-time_t
-mktime_from_utc(const struct tm *tm)
-{
-    int year = 1900 + tm->tm_year;
-    time_t days = 365 * (year - 1970) + leap_days(1970, year);
-    time_t hours;
-    time_t minutes;
-    time_t seconds;
-    int i;
-
-    for (i = 0; i < tm->tm_mon; ++i) {
-        days += mdays[i];
-    }
-    if (tm->tm_mon > 1 && is_leap_year(year)) {
-        ++days;
-    }
-    days += tm->tm_mday - 1;
-
-    hours = days * 24 + tm->tm_hour;
-    minutes = hours * 60 + tm->tm_min;
-    seconds = minutes * 60 + tm->tm_sec;
-
-    return seconds;
-}
-
 /* code to calculate CRC. Lifted from BSD 4.4 crc.c in cksum(1). BSD license.
    http://www.tsfr.org/~orc/Code/bsd/bsd-current/cksum/crc.c.
    or http://gobsd.com/code/freebsd/usr.bin/cksum/crc.c
    The polynomial is 0x04c11db7L. */
-static u_long crctab[] = {
+static uint32_t crctab[] = {
 	0x0,
 	0x04c11db7, 0x09823b6e, 0x0d4326d9, 0x130476dc, 0x17c56b6b,
 	0x1a864db2, 0x1e475005, 0x2608edb8, 0x22c9f00f, 0x2f8ad6d6,
@@ -855,9 +840,15 @@ compare_serial(uint32_t a, uint32_t b)
 uint16_t
 qid_generate(void)
 {
-#ifdef HAVE_ARC4RANDOM_UNIFORM
-    return (uint16_t) arc4random_uniform(65536);
-#elif HAVE_ARC4RANDOM
+#ifdef HAVE_GETRANDOM
+	uint16_t r;
+	if(getrandom(&r, sizeof(r), 0) == -1) {
+		log_msg(LOG_ERR, "getrandom failed: %s", strerror(errno));
+		exit(1);
+	}
+	return r;
+#elif defined(HAVE_ARC4RANDOM)
+    /* arc4random_uniform not needed because range is a power of 2 */
     return (uint16_t) arc4random();
 #else
     return (uint16_t) random();
@@ -867,9 +858,16 @@ qid_generate(void)
 int
 random_generate(int max)
 {
-#ifdef HAVE_ARC4RANDOM_UNIFORM
+#ifdef HAVE_GETRANDOM
+	int r;
+	if(getrandom(&r, sizeof(r), 0) == -1) {
+		log_msg(LOG_ERR, "getrandom failed: %s", strerror(errno));
+		exit(1);
+	}
+	return (int)(((unsigned)r)%max);
+#elif defined(HAVE_ARC4RANDOM_UNIFORM)
     return (int) arc4random_uniform(max);
-#elif HAVE_ARC4RANDOM
+#elif defined(HAVE_ARC4RANDOM)
     return (int) (arc4random() % max);
 #else
     return (int) ((unsigned)random() % max);
@@ -908,59 +906,63 @@ set_previous_owner(struct state_pretty_rr *state, const dname_type *dname)
 int
 print_rr(FILE *out,
          struct state_pretty_rr *state,
-         rr_type *record)
+         rr_type *record,
+	 region_type* rr_region,
+	 buffer_type* output)
 {
-	region_type *region = region_create(xalloc, free);
-        buffer_type *output = buffer_create(region, MAX_RDLENGTH);
-        rrtype_descriptor_type *descriptor
-                = rrtype_descriptor_by_type(record->type);
+        const nsd_type_descriptor_type *descriptor =
+		nsd_type_descriptor(record->type);
         int result;
         const dname_type *owner = domain_dname(record->owner);
-        const dname_type *owner_origin
-                = dname_origin(region, owner);
+	buffer_clear(output);
         if (state) {
-			if (!state->previous_owner
-				|| dname_compare(state->previous_owner, owner) != 0) {
-				int origin_changed = (!state->previous_owner_origin
-					|| dname_compare(state->previous_owner_origin,
-					   owner_origin) != 0);
-				if (origin_changed) {
-					buffer_printf(output, "$ORIGIN %s\n",
-						dname_to_string(owner_origin, NULL));
-				}
-
-				set_previous_owner(state, owner);
-				buffer_printf(output, "%s",
-					dname_to_string(owner,
-						state->previous_owner_origin));
+		if (!state->previous_owner
+			|| dname_compare(state->previous_owner, owner) != 0) {
+			const dname_type *owner_origin
+				= dname_origin(rr_region, owner);
+			int origin_changed = (!state->previous_owner_origin
+				|| dname_compare(state->previous_owner_origin,
+				   owner_origin) != 0);
+			if (origin_changed) {
+				buffer_printf(output, "$ORIGIN %s\n",
+					dname_to_string(owner_origin, NULL));
 			}
-		} else {
-			buffer_printf(output, "%s", dname_to_string(owner, NULL));
+
+			set_previous_owner(state, owner);
+			buffer_printf(output, "%s",
+				dname_to_string(owner,
+					state->previous_owner_origin));
+			region_free_all(rr_region);
 		}
-
-		buffer_printf(output, "\t%lu\t%s\t%s",
-			(unsigned long) record->ttl,
-			rrclass_to_string(record->klass),
-			rrtype_to_string(record->type));
-
-		result = print_rdata(output, descriptor, record);
-		if (!result) {
-			/*
-			 * Some RDATA failed to print, so print the record's
-			 * RDATA in unknown format.
-			 */
-			result = rdata_atoms_to_unknown_string(output,
-				descriptor, record->rdata_count, record->rdatas);
-		}
-
-		if (result) {
-			buffer_printf(output, "\n");
-			buffer_flip(output);
-			result = write_data(out, buffer_current(output),
-			buffer_remaining(output));
+	} else {
+		buffer_printf(output, "%s", dname_to_string(owner, NULL));
 	}
 
-	region_destroy(region);
+	buffer_printf(output, "\t%lu\t%s\t%s",
+		(unsigned long) record->ttl,
+		rrclass_to_string(record->klass),
+		rrtype_to_string(record->type));
+
+	if(record->type == TYPE_SOA) {
+		buffer_printf(output, "\t");
+		result = print_soa_rdata_twoline(output, record);
+	} else {
+		result = print_rdata(output, descriptor, record);
+	}
+	if (!result) {
+		/*
+		 * Some RDATA failed to print, so print the record's
+		 * RDATA in unknown format.
+		 */
+		result = print_unknown_rdata(output, descriptor, record);
+	}
+
+	if (result) {
+		buffer_printf(output, "\n");
+		buffer_flip(output);
+		result = write_data(out, buffer_current(output),
+		buffer_remaining(output));
+	}
 	return result;
 }
 
@@ -973,7 +975,7 @@ rcode2str(int rc)
 		case RCODE_FORMAT:
 			return "FORMAT ERROR";
 		case RCODE_SERVFAIL:
-			return "SERV FAIL";
+			return "SERVFAIL";
 		case RCODE_NXDOMAIN:
 			return "NAME ERROR";
 		case RCODE_IMPL:
@@ -989,6 +991,7 @@ rcode2str(int rc)
 		case RCODE_NOTAUTH:
 			return "SERVER NOT AUTHORITATIVE FOR ZONE";
 		case RCODE_NOTZONE:
+			/* Name not contained in zone */
 			return "NOTZONE";
 		default:
 			return "UNKNOWN ERROR";
@@ -1017,3 +1020,352 @@ addr2str(
 		str, len))
 		strlcpy(str, "[unknown ip4, inet_ntop failed]", len);
 }
+
+void
+addrport2str(
+#ifdef INET6
+	struct sockaddr_storage *addr
+#else
+	struct sockaddr_in *addr
+#endif
+	, char* str, size_t len)
+{
+	char ip[256];
+#ifdef INET6
+	if (addr->ss_family == AF_INET6) {
+		if (!inet_ntop(AF_INET6,
+			&((struct sockaddr_in6 *)addr)->sin6_addr, ip, sizeof(ip)))
+			strlcpy(ip, "[unknown ip6, inet_ntop failed]", sizeof(ip));
+		/* append port number */
+		snprintf(str, len, "%s@%u", ip,
+			(unsigned)ntohs(((struct sockaddr_in6 *)addr)->sin6_port));
+		return;
+	} else
+#endif
+	if (!inet_ntop(AF_INET, &((struct sockaddr_in *)addr)->sin_addr,
+		ip, sizeof(ip)))
+		strlcpy(ip, "[unknown ip4, inet_ntop failed]", sizeof(ip));
+	/* append port number */
+	snprintf(str, len, "%s@%u", ip,
+		(unsigned)ntohs(((struct sockaddr_in *)addr)->sin_port));
+}
+
+void
+append_trailing_slash(const char** dirname, region_type* region)
+{
+	int l = strlen(*dirname);
+	if (l>0 && (*dirname)[l-1] != '/' && l < 0xffffff) {
+		char *dirname_slash = region_alloc(region, l+2);
+		memcpy(dirname_slash, *dirname, l+1);
+		strlcat(dirname_slash, "/", l+2);
+		/* old dirname is leaked, this is only used for chroot, once */
+		*dirname = dirname_slash;
+	}
+}
+
+int
+file_inside_chroot(const char* fname, const char* chr)
+{
+	/* true if filename starts with chroot or is not absolute */
+	return ((fname && fname[0] && strncmp(fname, chr, strlen(chr)) == 0) ||
+		(fname && fname[0] != '/'));
+}
+
+/*
+ * Something went wrong, give error messages and exit.
+ */
+void
+error(const char *format, ...)
+{
+	va_list args;
+	va_start(args, format);
+	log_vmsg(LOG_ERR, format, args);
+	va_end(args);
+	exit(1);
+}
+
+#ifdef HAVE_CPUSET_T
+#if defined(HAVE_SYSCONF) && defined(_SC_NPROCESSORS_CONF)
+/* exists on Linux and FreeBSD */
+int number_of_cpus(void)
+{
+	return (int)sysconf(_SC_NPROCESSORS_CONF);
+}
+#else
+int number_of_cpus(void)
+{
+	return -1;
+}
+#endif
+#ifdef __gnu_hurd__
+/* HURD has no sched_setaffinity implementation, but links an always fail,
+ * with a linker error, we print an error when it is used */
+int set_cpu_affinity(cpuset_t *ATTR_UNUSED(set))
+{
+	log_msg(LOG_ERR, "sched_setaffinity: not available on this system");
+	return -1;
+}
+#elif defined(HAVE_SCHED_SETAFFINITY)
+/* Linux */
+int set_cpu_affinity(cpuset_t *set)
+{
+	assert(set != NULL);
+	return sched_setaffinity(getpid(), sizeof(*set), set);
+}
+#else
+/* FreeBSD */
+int set_cpu_affinity(cpuset_t *set)
+{
+	assert(set != NULL);
+	return cpuset_setaffinity(
+		CPU_LEVEL_WHICH, CPU_WHICH_PID, -1, sizeof(*set), set);
+}
+#endif
+#endif /* HAVE_CPUSET_T */
+
+void add_cookie_secret(struct nsd* nsd, uint8_t* secret)
+{
+	/* New cookie secret becomes the staging secret (position 1)
+	 * unless there is no active cookie yet, then it becomes the active
+	 * secret.  If the NSD_COOKIE_HISTORY_SIZE > 2 then all staging cookies
+	 * are moved one position down.
+	 */
+	if(nsd->cookie_count == 0) {
+		memcpy( nsd->cookie_secrets->cookie_secret
+		       , secret, NSD_COOKIE_SECRET_SIZE);
+		nsd->cookie_count = 1;
+		explicit_bzero(secret, NSD_COOKIE_SECRET_SIZE);
+		return;
+	}
+#if NSD_COOKIE_HISTORY_SIZE > 2
+	memmove( &nsd->cookie_secrets[2], &nsd->cookie_secrets[1]
+	       , sizeof(struct cookie_secret) * (NSD_COOKIE_HISTORY_SIZE - 2));
+#endif
+	memcpy( nsd->cookie_secrets[1].cookie_secret
+	      , secret, NSD_COOKIE_SECRET_SIZE);
+	nsd->cookie_count = nsd->cookie_count     < NSD_COOKIE_HISTORY_SIZE
+	                  ? nsd->cookie_count + 1 : NSD_COOKIE_HISTORY_SIZE;
+	explicit_bzero(secret, NSD_COOKIE_SECRET_SIZE);
+}
+
+void activate_cookie_secret(struct nsd* nsd)
+{
+	uint8_t active_secret[NSD_COOKIE_SECRET_SIZE];
+	/* The staging secret becomes the active secret.
+	 * The active secret becomes a staging secret.
+	 * If the NSD_COOKIE_HISTORY_SIZE > 2 then all staging secrets are moved
+	 * one position up and the previously active secret becomes the last
+	 * staging secret.
+	 */
+	if(nsd->cookie_count < 2)
+		return;
+	memcpy( active_secret, nsd->cookie_secrets[0].cookie_secret
+	      , NSD_COOKIE_SECRET_SIZE);
+	memmove( &nsd->cookie_secrets[0], &nsd->cookie_secrets[1]
+	       , sizeof(struct cookie_secret) * (NSD_COOKIE_HISTORY_SIZE - 1));
+	memcpy( nsd->cookie_secrets[nsd->cookie_count - 1].cookie_secret
+	      , active_secret, NSD_COOKIE_SECRET_SIZE);
+	explicit_bzero(active_secret, NSD_COOKIE_SECRET_SIZE);
+}
+
+void drop_cookie_secret(struct nsd* nsd)
+{
+	/* Drops a staging cookie secret. If there are more than one, it will
+	 * drop the last staging secret. */
+	if(nsd->cookie_count < 2)
+		return;
+	explicit_bzero( nsd->cookie_secrets[nsd->cookie_count - 1].cookie_secret
+	              , NSD_COOKIE_SECRET_SIZE);
+	nsd->cookie_count -= 1;
+}
+
+void reconfig_cookies(struct nsd* nsd, struct nsd_options* options)
+{
+	cookie_secret_type cookie_secrets[NSD_COOKIE_HISTORY_SIZE];
+	char secret[NSD_COOKIE_SECRET_SIZE * 2 + 2/*'\n' and '\0'*/];
+	FILE* f = NULL;
+	size_t count = 0;
+	const char* fn;
+	size_t i, j;
+
+	nsd->do_answer_cookie = options->answer_cookie;
+
+	/* Cookie secrets in the configuration file take precedence */
+	if(options->cookie_secret) {
+#ifndef NDEBUG
+		ssize_t len =
+#endif
+		hex_pton(options->cookie_secret,
+				nsd->cookie_secrets[0].cookie_secret,
+				NSD_COOKIE_SECRET_SIZE);
+		/* Cookie length guaranteed in configparser.y */
+		assert(len == NSD_COOKIE_SECRET_SIZE);
+		nsd->cookie_count = 1;
+		if(options->cookie_staging_secret) {
+#ifndef NDEBUG
+			len =
+#endif
+			hex_pton(options->cookie_staging_secret,
+					nsd->cookie_secrets[1].cookie_secret,
+					NSD_COOKIE_SECRET_SIZE);
+			/* Cookie length guaranteed in configparser.y */
+			assert(len == NSD_COOKIE_SECRET_SIZE);
+			nsd->cookie_count = 2;
+		}
+		/*************************************************************/
+		nsd->cookie_secrets_source = COOKIE_SECRETS_FROM_CONFIG;
+		return;
+		/*************************************************************/
+	}
+	/* Are cookies from file explicitly disabled? */
+	if(!(fn = nsd->options->cookie_secret_file))
+		goto generate_cookie_secrets;
+
+	else if((f = fopen(fn, "r")) != NULL)
+		; /* pass */
+
+	/* a non-existing cookie file is not necessarily an error */
+	else if(errno != ENOENT) {
+		log_msg( LOG_ERR
+		       , "error reading cookie secret file \"%s\": \"%s\""
+		       , fn, strerror(errno));
+		goto generate_cookie_secrets;
+
+	/* Only at startup cookie_secrets_source == COOKIE_SECRETS_NONE.
+	 * Only then the previous default file location will be tried
+	 * when the current default file location didn't exist.
+	 */
+	} else if(nsd->cookie_secrets_source == COOKIE_SECRETS_NONE
+	       && nsd->options->cookie_secret_file_is_default
+	       && (f = fopen((fn = CONFIGDIR"/nsd_cookiesecrets.txt"),"r")))
+		; /* pass */
+
+	else if(errno != ENOENT) {
+		log_msg( LOG_ERR
+		       , "error reading cookie secret file \"%s\": \"%s\""
+		       , fn, strerror(errno));
+		goto generate_cookie_secrets;
+	} else
+		goto generate_cookie_secrets;
+
+	/* cookie secret file exists and is readable */
+	for( count = 0; count < NSD_COOKIE_HISTORY_SIZE; count++ ) {
+		size_t secret_len = 0;
+		ssize_t decoded_len = 0;
+		if( fgets(secret, sizeof(secret), f) == NULL ) { break; }
+		secret_len = strlen(secret);
+		if( secret_len == 0 ) { break; }
+		assert( secret_len <= sizeof(secret) );
+		secret_len = secret[secret_len - 1] == '\n' ? secret_len - 1 : secret_len;
+		if( secret_len != NSD_COOKIE_SECRET_SIZE * 2 ) {
+			fclose(f);
+			log_msg( LOG_ERR
+			       , "error parsing cookie secret file \"%s\""
+			       , fn);
+			explicit_bzero(cookie_secrets, sizeof(cookie_secrets));
+			explicit_bzero(secret, sizeof(secret));
+			goto generate_cookie_secrets;
+		}
+		/* needed for `hex_pton`; stripping potential `\n` */
+		secret[secret_len] = '\0';
+		decoded_len = hex_pton(secret, cookie_secrets[count].cookie_secret,
+		                       NSD_COOKIE_SECRET_SIZE);
+		if( decoded_len != NSD_COOKIE_SECRET_SIZE ) {
+			fclose(f);
+			log_msg( LOG_ERR
+			       , "error parsing cookie secret file \"%s\""
+			       , fn);
+			explicit_bzero(cookie_secrets, sizeof(cookie_secrets));
+			explicit_bzero(secret, sizeof(secret));
+			goto generate_cookie_secrets;
+		}
+		explicit_bzero(secret, sizeof(secret));
+	}
+	fclose(f);
+	if(count) {
+		nsd->cookie_count = count;
+		memcpy(nsd->cookie_secrets, cookie_secrets, sizeof(cookie_secrets));
+		region_str_replace(  nsd->region
+		                  , &nsd->cookie_secrets_filename, fn );
+		explicit_bzero(cookie_secrets, sizeof(cookie_secrets));
+		/*************************************************************/
+		nsd->cookie_secrets_source = COOKIE_SECRETS_FROM_FILE;
+		return;
+		/*************************************************************/
+	}
+	explicit_bzero(cookie_secrets, sizeof(cookie_secrets));
+
+generate_cookie_secrets:
+	/* Calculate a new random secret */
+	srandom(getpid() ^ time(NULL));
+
+	for( j = 0; j < NSD_COOKIE_HISTORY_SIZE; j++) {
+#if defined(HAVE_SSL)
+		if (!RAND_status()
+		||  !RAND_bytes(nsd->cookie_secrets[j].cookie_secret, NSD_COOKIE_SECRET_SIZE))
+#endif
+		for (i = 0; i < NSD_COOKIE_SECRET_SIZE; i++)
+			nsd->cookie_secrets[j].cookie_secret[i] = random_generate(256);
+	}
+	nsd->cookie_count = 1;
+	/*********************************************************************/
+	nsd->cookie_secrets_source = COOKIE_SECRETS_GENERATED;
+	/*********************************************************************/
+}
+
+ssize_t
+print_socket_servers(struct nsd_bitset *bitset, char *buf, size_t bufsz)
+{
+	/* x and y are the start and end points of a range of set bits */
+	/* z is the last unset bit */
+	int i, x, y, z, n = (int)(bitset->size);
+	char *sep = "";
+	size_t off, written_total;
+	ssize_t written = 0;
+
+	assert(bufsz != 0);
+
+	off = written_total = 0;
+	x = y = z = -1;
+	for (i = 0; i <= n; i++) {
+		if (i == n || !nsd_bitset_isset(bitset, i)) {
+			written = 0;
+			if (i == n && x == -1) {
+				assert(y == -1);
+				assert(z == (n - 1));
+				written = snprintf(buf, bufsz, "(none)");
+			} else if (y > z) {
+				assert(x > z);
+				if (x == 0 && y == (n - 1)) {
+					assert(z == -1);
+					written = snprintf(buf+off, bufsz-off,
+						"(all)");
+				} else if (x == y) {
+					written = snprintf(buf+off, bufsz-off,
+						"%s%d", sep, x+1);
+				} else if (x == (y - 1)) {
+					written = snprintf(buf+off, bufsz-off,
+						"%s%d %d", sep, x+1, y+1);
+				} else {
+					assert(y > (x + 1));
+					written = snprintf(buf+off, bufsz-off,
+						"%s%d-%d", sep, x+1, y+1);
+				}
+			}
+			z = i;
+			if (written > 0) {
+				written_total += (size_t)written;
+				off = (written_total < bufsz) ? written_total : bufsz - 1;
+				sep = " ";
+			} else if (written < 0) {
+				return -1;
+			}
+		} else if (x <= z) {
+			x = y = i;
+		} else {
+			assert(x > z);
+			y = i;
+		}
+	}
+	return written_total;
+}
